#!/usr/bin/perl -w

use strict;

use Getopt::Std;

use DBI();
use Net::LDAP;

# command buffering of STDOUT
$| = 1;


my $prgname = "glite-dgas-convertDGAS2APEL.pl";

if ( !exists($ENV{GLITE_LOCATION}) || $ENV{GLITE_LOCATION} eq "" ) {
        $ENV{GLITE_LOCATION}="/opt/glite/";
	print "[DGAS2APEL] Environment variable GLITE_LOCATION not defined! Trying /opt/glite/ ...\n";
}

my $cfname = "glite-dgas-convertDGAS2APEL.conf";
my $conffile = "$ENV{GLITE_LOCATION}/etc/$cfname";



# APEL uses KBytes for LcgRecords.MemoryReal and LcgRecords.MemoryVirtual
# number in bits
my $memoryUnitAPEL = 1024*8;

# conversion in bits
my %memoryUnitsGGF = (
	 b => 1,
	 B => 8,
         k => (1024)*8, # this is for DGAS (old trans_in.log), not def. by GGF!
        kb => (1024)*8, # this is for DGAS (old trans_in.log), not def. by GGF!
	KB => (1024)*8,
        MB => (1024**2)*8,
        GB => (1024**3)*8,
        TB => (1024**4)*8,
        PB => (1024**5)*8,
        EB => (1024**6)*8,
        Kb => (1024),
        Mb => (1024**2),
        Gb => (1024**3),
        Tb => (1024**4),
        Pb => (1024**5),
        Eb => (1024**6),
);


# for EDG/LCG resources:
my $GRIS_TIMEOUT=15; # timeout of 15 seconds for LDAP queries.
my $GRIS_PORT = 2135;
my $GRIS_LDAPBASE = "mds-vo-name=local,o=grid";


# this is for different GLUE schema versions:
# make sure the attributes and appropriate filters correspond!
# (we don't use a single hash since the sequence would be unclear,
#  by using two hashes we can assure a certain sequence)
# IMPORTANT: attributes can be specified in two differnt ways on the
# GRIS:
# 1) ATTRIBUTE_NAME: <value>
#    Example:
#    GlueSiteName: INFN-T1
#    In this case specify the GRIS_..._ATTRIBUTE as "ATTRIBUTE_NAME":
#    Example: "type<N>" => "GlueSiteName"
# 2) ATTRIBUTE_NAME: ATTRIBUTE_EXTENSION=<value>
#    Example:
#    GlueForeignKey: GlueSiteUniqueID=INFN-T1
#    In this case specify the GRIS_..._ATTRIBUTE as
#    "ATTRIBUTE_NAME: ATTRIBUTE_EXTENSION":
#    Example: "type<N>" => "GlueForeignKey: GlueSiteUniqueID"
#

# filters and attributes for the site name (ExecutingSite)
my $NUM_GRIS_SITE_ATTRIBUTES = 3;
my %GRIS_SITE_FILTERS = (
			 "type1" => "(&(objectClass=GlueSite))",
			 "type2" => "(&(objectClass=SiteInfo))",
			 "type3" => "(&(objectClass=GlueTop))"
			 );
my %GRIS_SITE_ATTRIBUTES = (
			    "type1" => "GlueSiteName",
			    "type2" => "siteName",
			    "type3" => "GlueForeignKey: GlueSiteUniqueID"
			    );

# filters and attributes for the benchmarks (SpecInt2000 and SpecFloat2000):
my $NUM_GRIS_SI00_ATTRIBUTES = 2;
my %GRIS_SI00_FILTERS = (
			 "type1" => "(&(objectClass=GlueHostBenchmark))",
                         "type2" => "(&(objectClass=GlueHostBenchmark))"
 			 );
my %GRIS_SI00_ATTRIBUTES = (
			    "type1" => "GlueHostApplicationSoftwareRunTimeEnvironment: SI00MeanPerCPU",
			    #"type1" => "GlueHostApplicationSoftwareRunTimeEnvironment: SI00Mean",
			    "type2" => "GlueHostBenchmarkSI00"
			    );
my $NUM_GRIS_SF00_ATTRIBUTES = 2;
my %GRIS_SF00_FILTERS = (
			 "type1" => "(&(objectClass=GlueHostBenchmark))",
			 "type2" => "(&(objectClass=GlueHostBenchmark))"
			 );
my %GRIS_SF00_ATTRIBUTES = (
                            "type1" => "GlueHostApplicationSoftwareRunTimeEnvironment: SF00MeanPerCPU",
			    #"type1" => "GlueHostApplicationSoftwareRunTimeEnvironment: SF00Mean",
			    "type2" => "GlueHostBenchmarkSF00"
			    );



# for getValueFromGRIS() and getValueFromGRISMultipleAttributes():
my $GRIS_ERROR = "ERROR FROM GRIS";


# -------------------
# getting options ...
# -------------------

my %opts;
$opts{h} = "";
$opts{c} = "";
$opts{s} = "";
$opts{e} = "";
$opts{n} = "";
$opts{T} = "";
$opts{M} = "";
$opts{R} = "";
$opts{v} = "";

# specified on command line:
my $StartAtHLRtid = -1;
my $EndAtHLRtid = -1;
# determined during conversion (for entry in ConversionLog):
my $firstTIDconverted = -1;
my $lastTIDconverted = -1;


# default: write detailed log information to MAP_tid_RecordIdentity
my $detailedLog = 1;
my $verboseOutput = 0;

# retry to insert LcgRecords entries with error conditions from previous
# executions (table ConversionErrors)
my $retryErrors = 0;

# query the GRIS (for EDG resources, "-T edg") to retrieve the ExecutingSite
my $queryGRIS = 0;

# for each transaction the following value will be reset to 1, if a GRIS
# is contacted, but a timeout received, then it is set to 0, in order to
# prevent from trying to contact the same GRIS multiple times for the same
# transaction (which massively slows down the execution if GRISes temporarily
# not available):
my $GRISavailable = 1;

# map CEIDs to ExecutingSite names and benchmarks and use the mapping if the
# values cannot be retrieved in any other way
my $useMapping = 0;

getopts('hc:s:e:nT:MRv', \%opts);


if ( $opts{h} eq "1" ) {
    &printUsageInfo();
}

if ( $opts{c} ne "" ) {
    $conffile = $opts{c};
}

if ( $opts{s} ne "" ) {
    $StartAtHLRtid = $opts{s};
}

if ( $opts{e} ne "" ) {
    $EndAtHLRtid = $opts{e};
}

if ( $opts{n} eq "1" ) {
    $detailedLog = 0;
}

if ( $opts{T} eq "edg" ) {
    $queryGRIS = 1;
}

if ( $opts{M} eq "1" ) {
    $useMapping = 1;
}

if ( $opts{R} eq "1" ) {
    $retryErrors = 1;
}

if ( $opts{v} eq "1" ) {
    $verboseOutput = 1;
}


# ------------------------------
# parsing of configuration file:
# ------------------------------

# defaults for APEL conversion db:
# This is the database where the APEL LcgRecords are written
my $apelDbServer = "localhost";
my $apelDbPort = 3306;
my $apelDbUser = "root";
my $apelDbPassword = "";
my $apelDbName = "dgas2apel";

# defaults for conversion log db:
# This is the database where the conversion log is written to keep track of
# already converted HLR transactions
my $convDbServer = "localhost";
my $convDbPort = 3306;
my $convDbUser = "root";
my $convDbPassword = "";
my $convDbName = "dgas2apellog";

# defaults for DGAS HLR db:
# This is the database where the original HLR accounting information is taken
# from.
my $hlrDbServer = "localhost";
my $hlrDbPort = 3306;
my $hlrDbUser = "root";
my $hlrDbPassword = "";
my $hlrDbName = "hlr";

# additional info for LcgRecords table and conversion log entries: 
my $HLRLocation = "";

# Number of retries per transaction, before giving up (one retry per execution
# of the covnersion tool). Used only if -R specified. Default is 10.
my $maxNumRetries = 10;

# read configuration from file:
&readConfiguration();


# -----------------------
# connecting to MySQL dbs
# -----------------------

my $apelDbh;
my $convDbh;
my $hlrDbh;

my $apelDbQuery;
my $convDbQuery;
my $hlrDbQuery;

my $apelSth;
my $convSth;
my $hlrSth;
my $row;

&connectToDatabases();


# ------------------------------
# starting conversion procedure:
# ------------------------------

# will contain HLR transaction log (transInLog.log) entries:
my %logEntries = ();

# will contain converted info for LcgRecords:
my %lcgRecord = ();


my $exitStatus = 0; # 0 if ok


# -------------------------------------------------------------------------
# processing of transactions with error conditions (from previous runs) ...
# -------------------------------------------------------------------------

if ($retryErrors) { # only if -R specified
    $exitStatus += &retryPrevErrTransactions();  # add error status.
}


# --------------------------------------------------------------
# processing regular transactions from start tid to stop tid ...
# --------------------------------------------------------------

$exitStatus += &sequentialTransactionProcessing();  # add error status.



# ------------------------------------------
# finished conversion procedure, exiting ...
# ------------------------------------------

exit ($exitStatus);




# ===========================================================================
#                               FUNCTIONS
# ===========================================================================


# ---------------------------------------------------------------------------
#
# For retrying previous transactions with error conditions (ConversionErrors)
# arguments: none
# returns 0 if ok, otherwise 1
#
# ---------------------------------------------------------------------------
sub retryPrevErrTransactions() {

    my $returnStatus = 0; # ok

    print "[DGAS2APEL] Retrying to process transactions with error conditions from\n";
    print "[DGAS2APEL] previous executions of $prgname:\n";


    # checking for which HLR tids (transaction ids) to retry and how many
    # retries have already been done:
    my %tids2NumRetries = &getTransactionsWithConversionErrors();

    my @errorTids = keys(%tids2NumRetries);

    if (scalar(@errorTids) == 0) {
	print "[DGAS2APEL] No HLR transactions with error conditions from previous runs!\n";
	return 0; # nothing to do, but ok
    }


    # parse single transactions and insert them into the conversion DB:
    # determine start time of conversion:
    my $convStartTime = &timestamp2String(time(), "Z");
    my $tPos = index($convStartTime, "T");
    substr($convStartTime, $tPos, 1) = " ";
    my $zPos = index($convStartTime, "Z");
    substr($convStartTime, $zPos, 1) = "";

    print "[DGAS2APEL] Starting retries at $convStartTime UTC.\n";

    # ----------------------------------
    # now getting these transactions ...
    # ----------------------------------

    # Retrieved data per transaction:
    #
    # trans_in.tid          -> transID            = HLR transaction id
    # trans_in.dg_jobid     -> jobID              = global job ID
    # acctdesc.cert_subject -> resID              = global CE ID
    # transInLog.log        -> transLog    = transaction log with metrics, etc.
    # trans_in.tr_stamp     -> transTimestamp     = timestamp for transaction

    # Note: the info is distributed in 3 tables with the following relations:
    #   acctdesc.id = trans_in.rid
    #   trans_in.dg_jobid = transInLog.dgJobId
    # Furthermore we consider only transaction for the resources registered on
    # the HLR, NOT for the users that eventually are registered:
    #   acctdesc.a_type = 'resource'

    my $errTid;
    my $numRet;
    foreach $errTid (@errorTids) {

	$numRet = $tids2NumRetries{$errTid};
	print "[DGAS2APEL] Found transaction ID ".$errTid." with ".$numRet." retries:\n";

	# determine whether this is the last retry and insertion has to be
	# forced!
	my $forceInsertion = 0;
	if ($numRet+1 >= $maxNumRetries) {
	    # this is the last retry, since numRet + this retry reaches the max
	    print "[DGAS2APEL] ATTENTION: maximum number of retries reached, insertion will be\n";
	    print "[DGAS2APEL]            forced (if possible), even for an incomplete record!\n";
	    $forceInsertion = 1;
	}

	# getting the single transaction from HLR db:
	$hlrDbQuery = "SELECT trans_in.tid AS transID, trans_in.dg_jobid AS jobID, acctdesc.cert_subject AS resID, transInLog.log AS transLog, trans_in.tr_stamp AS transTimestamp FROM trans_in, acctdesc, transInLog WHERE acctdesc.a_type = 'resource' AND acctdesc.id = trans_in.rid AND trans_in.dg_jobid = transInLog.dgJobId AND trans_in.tid = $errTid";

	print "[DGAS2APEL] Getting transaction $errTid:\n[DGAS2APEL] Query = $hlrDbQuery\n";

	eval { $hlrSth = $hlrDbh->prepare($hlrDbQuery) };
	if ($@) {
	    print STDERR "[DGAS2APEL] Query preparation failed! Error: $@\n";
	    return 1; # error
	}

	eval { $hlrSth->execute() };
	if ($@) {
	    print STDERR "[DGAS2APEL] Query execution failed! Error: $@\n";
	    return 1; # error
	}

	if ($row = $hlrSth->fetchrow_hashref()) {

	    print "[DGAS2APEL] transaction: trans_in.tid=$row->{'transID'}\n";
	    print "[DGAS2APEL]              trans_in.dg_jobid=$row->{'jobID'}\n";
	    print "[DGAS2APEL]              acctdesc.cert_subject=$row->{'resID'}\n";
	    print "[DGAS2APEL]              transInLog.log=$row->{'transLog'}\n";
	    print "[DGAS2APEL]              trans_in.tr_stamp=$row->{'transTimestamp'}\n";

	    # parsing transaction log:
	    &getTransInLogEntries();


	    # contructing the APEL (LcgRecords) record:
	    my $recordOk = &constructAPELrecord();

	    my $recordInserted = 0;

	    if ($recordOk || $forceInsertion) {
		# insert the LcgRecords entry to the APEL database:
		$recordInserted = &insertAPELrecord();

		# if detailed log information is required and the LcgRecords entry was
		# successfully inserted in the APEL database: map tid to RecordIdentity
		if ($recordInserted && $detailedLog) {
		    &addTid2RecordIdentityMAP();
		}
	    }

	    if (!$recordInserted) {
		# not inserted, add to ConversionErrors (or increment the count)
		print "[DGAS2APEL] Record not inserted into APEL database, setting retry counter.\n";
		&incrementConversionErrorCount(int($row->{'transID'}));
	    }
	    else {
		# inserted, remove from ConversionErrors!
		print "[DGAS2APEL] Record inserted into APEL database, removing retry counter.\n";
		if (!&removeConversionError($errTid)) {
		    print "[DGAS2APEL] ERROR: Could not remove ConversionErrors entry for HLR_tid=$errTid\n";
		    print "[DGAS2APEL]        although the record was inserted into the APEL database!\n";
		    $returnStatus = 1;
		}
	    }

	} #if ($row ...

    } # foreach $errTid ...


    # determine stop time of conversion:
    my $convStopTime = &timestamp2String(time(), "Z");
    $tPos = index($convStopTime, "T");
    substr($convStopTime, $tPos, 1) = " ";
    $zPos = index($convStopTime, "Z");
    substr($convStopTime, $zPos, 1) = "";

    print "[DGAS2APEL] Stopping retries at $convStopTime UTC.\n";

    return $returnStatus; # 0 is ok, 1 is error

} # retryPrevErrTransactions() ...



# ---------------------------------------------------------------------------
#
# For the regular sequential transaction processing from start tid to stop
# tid.
# arguments: none
# returns 0 if ok, otherwise 1
#
# ---------------------------------------------------------------------------
sub sequentialTransactionProcessing() {

    print "[DGAS2APEL] Sequential processing of transactions:\n";

    # checking for which HLR tids (transaction ids) to convert:
    &determineTidRange();

    print "[DGAS2APEL] Starting at HLR tid (transaction id): $StartAtHLRtid\n";
    print "[DGAS2APEL] Ending at HLR tid (transaction id): $EndAtHLRtid\n";

    if ($EndAtHLRtid < $StartAtHLRtid) {
	print "[DGAS2APEL] No HLR transactions to convert to LcgRecords entries!\n";
	return 0; # nothing to do, but ok
    }



    # parse single transactions and insert them into the conversion DB:
    # determine start time of conversion:
    my $convStartTime = &timestamp2String(time(), "Z");
    my $tPos = index($convStartTime, "T");
    substr($convStartTime, $tPos, 1) = " ";
    my $zPos = index($convStartTime, "Z");
    substr($convStartTime, $zPos, 1) = "";

    print "[DGAS2APEL] Starting sequential conversion at $convStartTime UTC.\n";

    # ----------------------------------
    # now getting these transactions ...
    # ----------------------------------

    # Retrieved data per transaction:
    #
    # trans_in.tid          -> transID            = HLR transaction id
    # trans_in.dg_jobid     -> jobID              = global job ID
    # acctdesc.cert_subject -> resID              = global CE ID
    # transInLog.log        -> transLog    = transaction log with metrics, etc.
    # trans_in.tr_stamp     -> transTimestamp     = timestamp for transaction

    # Note: the info is distributed in 3 tables with the following relations:
    #   acctdesc.id = trans_in.rid
    #   trans_in.dg_jobid = transInLog.dgJobId
    # Furthermore we consider only transaction for the resources registered on
    # the HLR, NOT for the users that eventually are registered:
    #   acctdesc.a_type = 'resource'

    $hlrDbQuery = "SELECT trans_in.tid AS transID, trans_in.dg_jobid AS jobID, acctdesc.cert_subject AS resID, transInLog.log AS transLog, trans_in.tr_stamp AS transTimestamp FROM trans_in, acctdesc, transInLog WHERE acctdesc.a_type = 'resource' AND acctdesc.id = trans_in.rid AND trans_in.dg_jobid = transInLog.dgJobId AND trans_in.tid >= $StartAtHLRtid AND trans_in.tid <= $EndAtHLRtid";

    print "[DGAS2APEL] Getting transactions:\n[DGAS2APEL] Query = $hlrDbQuery\n";

    eval { $hlrSth = $hlrDbh->prepare($hlrDbQuery) };
    if ($@) {
	print STDERR "[DGAS2APEL] Query preparation failed! Error: $@\n";
	return 1; # error
    }

    eval { $hlrSth->execute() };
    if ($@) {
	print STDERR "[DGAS2APEL] Query execution failed! Error: $@\n";
	return 1; # error
    }


    while ($row = $hlrSth->fetchrow_hashref()) {

	print "[DGAS2APEL] transaction: trans_in.tid=$row->{'transID'}\n";
	print "[DGAS2APEL]              trans_in.dg_jobid=$row->{'jobID'}\n";
	print "[DGAS2APEL]              acctdesc.cert_subject=$row->{'resID'}\n";
	print "[DGAS2APEL]              transInLog.log=$row->{'transLog'}\n";
	print "[DGAS2APEL]              trans_in.tr_stamp=$row->{'transTimestamp'}\n";

	# parsing transaction log:
	&getTransInLogEntries();


	# contructing the APEL (LcgRecords) record:
	my $recordOk = &constructAPELrecord();

	my $recordInserted = 0;
	if ($recordOk) {
	    # insert the LcgRecords entry to the APEL database:
	    $recordInserted = &insertAPELrecord();

	    # if detailed log information is required and the LcgRecords entry was
	    # successfully inserted in the APEL database: map tid to RecordIdentity
	    if ($recordInserted && $detailedLog) {
		&addTid2RecordIdentityMAP();
	    }
	}


	if (!$recordInserted && $retryErrors) {
	    # not inserted, add to ConversionErrors (or increment the count)
	    print "[DGAS2APEL] Record not inserted into APEL database, setting retry counter.\n";
	    &incrementConversionErrorCount(int($row->{'transID'}));
	}


	# This is for the ConversionLog:
	# determine first and last converted transaction:
	if ($firstTIDconverted == -1) {
	    # first transaction
	    $firstTIDconverted = $row->{'transID'};
	}
	# always: this might be the last transaction
	$lastTIDconverted = $row->{'transID'};

    } # while row (transaction)



    # determine stop time of conversion:
    my $convStopTime = &timestamp2String(time(), "Z");
    $tPos = index($convStopTime, "T");
    substr($convStopTime, $tPos, 1) = " ";
    $zPos = index($convStopTime, "Z");
    substr($convStopTime, $zPos, 1) = "";

    print "[DGAS2APEL] Stopping sequential conversion at $convStopTime UTC.\n";


    # write an entry to the summarized conversion log in ConversionLog
    $convDbQuery = "INSERT INTO ConversionLog VALUES ("
	."NULL,"
	."\"$convStartTime\","
	."\"$convStopTime\","
	."\"$HLRLocation\","
	."$firstTIDconverted,"
	."$lastTIDconverted"
	.")";
    # check which values are NULL and do not specify them as a string
    # (that is, relpace >"NULL"< with >NULL<).
    $convDbQuery =~ s/"NULL"/NULL/g;

    # Insert the start and end tid info into the conversion log db:
    print "[DGAS2APEL] Insert summarized log (start and end tid) in ConversionLog:\n[DGAS2APEL] Query = $convDbQuery\n";

    eval { $convSth = $convDbh->prepare($convDbQuery) };
    if ($@) {
	print STDERR "[DGAS2APEL] Query preparation failed! Error: $@\n";
	return 1; # error
    }

    eval { $convSth->execute() };
    if ($@) {
	print STDERR "[DGAS2APEL] Query execution failed! Error: $@\n";
	return 1; # error
    }

    return 0; # ok

} # sequentialTransactionProcessing() ...




# ---------------------------------------------------------------------------
#
# arguments: timestamp, timeZone(e.g. "+0200")
# returns a local time if the second argument is not "+0000" or "Z"
#
# ---------------------------------------------------------------------------
sub timestamp2String () {

    my $tstamp = $_[0];
    my $tzStr = $_[1];

    # parse timeZone:
    if ($tzStr eq "+0000" || $tzStr eq "Z") {
	$tzStr = "Z";
    }
    elsif ($tzStr =~ /^(\+|-)\d{4}/) {
	my $hourDiff;
	my $minDiff;

	$hourDiff = int(substr($tzStr,0,3));
	$minDiff = int(substr($tzStr,3,2));

#	if ($hourDiff < 0) { <- this is wrong since it misses e.g. -0030 !
	if ($tzStr =~ /^-/) {
	    # negative time zone, consider minutes negative as well!
	    $minDiff = $minDiff * (-1);
	}

	# compute difference to UTC:
	$tstamp = $tstamp + ($hourDiff * 3600) + ($minDiff * 60);
    }
    else {
	if ($verboseOutput) {
	    print "[DGAS2APEL]              Wrong timezone format: $tzStr - cannot convert timestamp\n";
	}
	return "NULL";
    }

    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst);

    eval {
	($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = gmtime($tstamp);
    };
    if ($@) {
	if ($verboseOutput) {
	    print "[DGAS2APEL]              Wrong timestamp format: $tstamp - cannot convert timestamp\n";
	}
	return "NULL";
    }
    
    my $sTime = "";
    # year
    $year = $year+1900;
    # month
    $mon = $mon+1;
    # day

    # hour
    $hour = $hour;
    # minute
    $min = $min;
    # second
    $sec = $sec;


    # composing string:
    $sTime = "".$year."-";
    if ($mon < 10) {
	$sTime = $sTime."0".$mon."-";
    }
    else {
	$sTime = $sTime.$mon."-";
    }
    if ($mday < 10) {
	$sTime = $sTime."0".$mday."T";
    }
    else {
	$sTime = $sTime.$mday."T";
    }
    if ($hour < 10) {
	$sTime = $sTime."0".$hour.":";
    }
    else {
	$sTime = $sTime.$hour.":";
    }
    if ($min < 10) {
	$sTime = $sTime."0".$min.":";
    }
    else {
	$sTime = $sTime.$min.":";
    }
    if ($sec < 10) {
	$sTime = $sTime."0".$sec;
    }
    else {
	$sTime = $sTime.$sec;
    }

    return $sTime.$tzStr;
}


# ---------------------------------------------------------------------------
#
# argument: time in seconds
# returns a phase string (time period) for a time in seconds
#
# ---------------------------------------------------------------------------
sub timeSecs2Period () {

    # compute it in the human readable format (as phase units):
    # Since the length of a month is unclear we prefer not to use it ...
    # according to the GGF UR-WG Usage Record - Format Recomendation
    # (draft of March 2005): "... only resources that are typically
    #  measured in month and year intervals should use that component
    #  of duration." (e.g. software licenses, etc.)

    my $tSecs = $_[0];

    my $timePeriodString = "P";

    if ($tSecs =~ /^\d*$/) {

	my $remainingSecs = $tSecs;
	my $days = int($tSecs/86400);
	$remainingSecs = $remainingSecs - ($days * 86400);
	my $hours = int($remainingSecs/3600);
	$remainingSecs = $remainingSecs - ($hours * 3600);
	my $minutes = int($remainingSecs/60);
	$remainingSecs = $remainingSecs - ($minutes * 60);
	#print "days=$days; hours=$hours; minutes=$minutes; remainingSecs=$remainingSecs\n";
	if ($days > 0) {
	    $timePeriodString = $timePeriodString."$days"."DT";
	}
	if ($hours > 0) {
	    $timePeriodString = $timePeriodString."$hours"."H";
	}
	if ($minutes > 0) {
	    $timePeriodString = $timePeriodString."$minutes"."M";
	}
	if ($remainingSecs > 0) {
	    $timePeriodString = $timePeriodString."$remainingSecs"."S";
	}

    }
    else {
	if ($verboseOutput) {
	    print "[DGAS2APEL]              Wrong time(seconds) format: $tSecs - cannot convert to time period\n";
	}
	return "NULL";
    }

    return $timePeriodString;
}



# ---------------------------------------------------------------------------
#
# read configuration file and parse parameters:
# arguments: none
#
# ---------------------------------------------------------------------------
sub readConfiguration() {

    # first read the file:
    my $line = "";

    unless ( -e $conffile ) {
	print STDERR "[DGAS2APEL] Error: configuration file $conffile not found!\n";
	exit(1);
    }

    unless ( -r $conffile ) {
	print STDERR "[DGAS2APEL] Error: configuration file $conffile not readable!\n";
	exit(1);
    }
         
    unless ( open(IN, "< $conffile") ) {
	print STDERR "[DGAS2APEL] Error: cannot open configuration file $conffile!\n";
        exit(1);
    }

    my %confparameters;

    # reading the single lines
    while ($line = <IN>) {
	chomp($line);

	$line || next;

	next if ($line =~ /^\s*\#+/);

	if ($line =~ /(.*)\s*=\s*\"(.*)\"/) {

	    $confparameters{$1} = $2;

	}

    }

    #my $confpar;
    #foreach $confpar (keys(%confparameters)) {
    #    print "$confpar = $confparameters{$confpar}\n";
    #}

    # close the configuration file
    close(IN);


    # now parse the parameters:

    # get APEL conversion db info:

    if (exists($confparameters{"APEL_DB_SERVER"})) {
	$apelDbServer = $confparameters{"APEL_DB_SERVER"};
    }
    else {
	print "[DGAS2APEL] Configuration parameter APEL_DB_SERVER missing, trying '$apelDbServer'!\n";
    }

    if (exists($confparameters{"APEL_DB_PORT"})) {
	$apelDbPort = $confparameters{"APEL_DB_PORT"};
    }
    else {
	print "[DGAS2APEL] Configuration parameter APEL_DB_PORT missing, trying '$apelDbPort'!\n";
    }

    if (exists($confparameters{"APEL_DB_USER"})) {
	$apelDbUser = $confparameters{"APEL_DB_USER"};
    }
    else {
	print "[DGAS2APEL] Configuration parameter APEL_DB_USER missing, trying '$apelDbUser'!\n";
    }

    if (exists($confparameters{"APEL_DB_PASSWORD"})) {
	$apelDbPassword = $confparameters{"APEL_DB_PASSWORD"};
    }
    else {
	print "[DGAS2APEL] Configuration parameter APEL_DB_PASSWORD missing, trying '$apelDbPassword'!\n";
    }

    if (exists($confparameters{"APEL_DB_NAME"})) {
	$apelDbName = $confparameters{"APEL_DB_NAME"};
    }
    else {
	print "[DGAS2APEL] Configuration parameter APEL_DB_NAME missing, trying '$apelDbName'!\n";
    }

    # get conversion log db info:

    if (exists($confparameters{"CONV_DB_SERVER"})) {
	$convDbServer = $confparameters{"CONV_DB_SERVER"};
    }
    else {
	print "[DGAS2APEL] Configuration parameter CONV_DB_SERVER missing, trying '$convDbServer'!\n";
    }

    if (exists($confparameters{"CONV_DB_PORT"})) {
	$convDbPort = $confparameters{"CONV_DB_PORT"};
    }
    else {
	print "[DGAS2APEL] Configuration parameter CONV_DB_PORT missing, trying '$convDbPort'!\n";
    }

    if (exists($confparameters{"CONV_DB_USER"})) {
	$convDbUser = $confparameters{"CONV_DB_USER"};
    }
    else {
	print "[DGAS2APEL] Configuration parameter CONV_DB_USER missing, trying '$convDbUser'!\n";
    }

    if (exists($confparameters{"CONV_DB_PASSWORD"})) {
	$convDbPassword = $confparameters{"CONV_DB_PASSWORD"};
    }
    else {
	print "[DGAS2APEL] Configuration parameter CONV_DB_PASSWORD missing, trying '$convDbPassword'!\n";
    }

    if (exists($confparameters{"CONV_DB_NAME"})) {
	$convDbName = $confparameters{"CONV_DB_NAME"};
    }
    else {
	print "[DGAS2APEL] Configuration parameter CONV_DB_NAME missing, trying '$convDbName'!\n";
    }

    # get DGAS HLR db info:

    if (exists($confparameters{"HLR_DB_SERVER"})) {
	$hlrDbServer = $confparameters{"HLR_DB_SERVER"};
    }
    else {
	print "[DGAS2APEL] Configuration parameter HLR_DB_SERVER missing, trying '$hlrDbServer'!\n";
    }

    if (exists($confparameters{"HLR_DB_PORT"})) {
	$hlrDbPort = $confparameters{"HLR_DB_PORT"};
    }
    else {
	print "[DGAS2APEL] Configuration parameter HLR_DB_PORT missing, trying '$hlrDbPort'!\n";
    }

    if (exists($confparameters{"HLR_DB_USER"})) {
	$hlrDbUser = $confparameters{"HLR_DB_USER"};
    }
    else {
	print "[DGAS2APEL] Configuration parameter HLR_DB_USER missing, trying '$hlrDbUser'!\n";
    }

    if (exists($confparameters{"HLR_DB_PASSWORD"})) {
	$hlrDbPassword = $confparameters{"HLR_DB_PASSWORD"};
    }
    else {
	print "[DGAS2APEL] Configuration parameter HLR_DB_PASSWORD missing, trying '$hlrDbPassword'!\n";
    }

    if (exists($confparameters{"HLR_DB_NAME"})) {
	$hlrDbName = $confparameters{"HLR_DB_NAME"};
    }
    else {
	print "[DGAS2APEL] Configuration parameter HLR_DB_NAME missing, trying '$hlrDbName'!\n";
    }

    # get additional info for LcgRecords
    if (exists($confparameters{"HLR_LOCATION"})) {
	$HLRLocation = $confparameters{"HLR_LOCATION"};
    }
    else {
	print STDERR "[DGAS2APEL] Error: Configuration parameter HLR_LOCATION missing!\n";
	print STDERR "[DGAS2APEL] This parameter is required and has to be of the form:\n";
	print STDERR "[DGAS2APEL]  '<host>:<port>:<subject of host certificate>'\n";
	print STDERR "[DGAS2APEL] Example: grid005.to.infn.it:56568:/C=IT/O=INFN/OU=Host/L=Torino/CN=grid005.to.infn.it\n";
	exit(1);
    }

    # get maximum number of retries (used if -R specified)
    if (exists($confparameters{"MAX_NUM_RETRIES"})) {
	$maxNumRetries = $confparameters{"MAX_NUM_RETRIES"};
    }
    else {
	print "[DGAS2APEL] Configuration parameter MAX_NUM_RETRIES missing, default is ".$maxNumRetries.".\n";
    }


} # readConfiguration() ...



# ---------------------------------------------------------------------------
#
# connecting to the databases (APEL, DGAS, ConversionLog)
# argument: none
#
# ---------------------------------------------------------------------------
sub connectToDatabases() {

    # APEL conversion db:

    eval { $apelDbh = DBI->connect("DBI:mysql:database=$apelDbName;host=$apelDbServer",
				   "$apelDbUser", "$apelDbPassword",
				   {'RaiseError' => 1}
				   )
	   };
    if ($@) {
	print STDERR "[DGAS2APEL] Error: Cannot connect to APEL conversion database! Error: $@\n";
	exit(1);
    }

    # conversion log db:

    eval { $convDbh = DBI->connect("DBI:mysql:database=$convDbName;host=$convDbServer",
				   "$convDbUser", "$convDbPassword",
				   {'RaiseError' => 1}
				   )
	   };
    if ($@) {
	print STDERR "[DGAS2APEL] Error: Cannot connect to conversion log database! Error: $@\n";
	exit(1);
    }

    # DGAS HLR db:

    eval { $hlrDbh = DBI->connect("DBI:mysql:database=$hlrDbName;host=$hlrDbServer",
				  "$hlrDbUser", "$hlrDbPassword",
				  {'RaiseError' => 1}
				  )
	   };
    if ($@) {
	print STDERR "[DGAS2APEL] Error: Cannot connect to DGAS HLR database! Error: $@\n";
	exit(1);
    }

} # connectToDatabases() ...


# ---------------------------------------------------------------------------
#
# determine range of transaction IDs (tid) to be processed.
# arguments: none
#
# ---------------------------------------------------------------------------
sub determineTidRange() {

    # get max tid and adding 1 or setting to 0 (no conversions so far):

    if ($StartAtHLRtid < 0) {

	$convDbQuery = "SELECT max(To_HLR_tid) FROM ConversionLog WHERE HLRLocation=\"".$HLRLocation."\"";

	eval { $convSth = $convDbh->prepare($convDbQuery) };
	if ($@) {
	    print STDERR "[DGAS2APEL] Query preparation failed! Query: '$convDbQuery';\nError: $@\n";
	    exit(1);
	}

	eval { $convSth->execute() };
	if ($@) {
	    print STDERR "[DGAS2APEL] Query execution failed! Query: '$convDbQuery';\nError: $@\n";
	    exit(1);
	}

	$row = $convSth->fetchrow_hashref();
	if ($row->{'max(To_HLR_tid)'}) {
	    $StartAtHLRtid = $row->{'max(To_HLR_tid)'} + 1;
	}
	else {
	    $StartAtHLRtid = 0;
	}
    }

    # get max tid and add 1 --- or setting to 0 (no conversions so far):

    if ($EndAtHLRtid < 0) {

	$hlrDbQuery = "SELECT max(tid) FROM trans_in";

	eval { $hlrSth = $hlrDbh->prepare($hlrDbQuery) };
	if ($@) {
	    print STDERR "[DGAS2APEL] Query preparation failed! Query: '$hlrDbQuery';\nError: $@\n";
	    exit(1);
	}

	eval { $hlrSth->execute() };
	if ($@) {
	    print STDERR "[DGAS2APEL] Query execution failed! Query: '$hlrDbQuery';\nError: $@\n";
	    exit(1);
	}

	$row = $hlrSth->fetchrow_hashref();
	if ($row->{'max(tid)'}) {
	    $EndAtHLRtid = $row->{'max(tid)'};
	}
	else {
	    $EndAtHLRtid = 0;
	}
    }

} # determineTidRange() ...



# ---------------------------------------------------------------------------
#
# Print usage information (help).
# arguments: none
#
# ---------------------------------------------------------------------------
sub printUsageInfo() {
    print "Author: Rosario Piro (piro\@to.infn.it)\n\n";
    print "Usage:\n";
    print "$prgname [OPTIONS]\n";
    print "\nWhere OPTIONS are:\n";
    print "-c <file>    Specifies the configuration file.\n";
    print "             Default: \$GLITE_LOCATION/etc/$cfname\n";
    print "-s <tid>     HLR transaction id from which to start the conversion\n";
    print "             Default: Determined from the last entry in the ConversionLog\n";
    print "             table: start point = highest tid (converted) + 1\n";
    print "-e <tid>     HLR transaction id up to which to convert\n";
    print "             Default: Determined from the HLR database:\n";
    print "             end point = highest tid found in the trans_in table\n";
    print "-n           No detailed log information in table MAP_tid_RecordIdentity\n";
    print "             (summarized log information in ConversionLog will always\n";
    print "             be written).\n";
    print "-T <ceType>  Type of CEs for which accounting information is to be converted.\n";
    print "             Possible values for <ceType>:\n";
    print "             \"edg\": a GRIS can be contacted for each CE to retrieve some\n";
    print "                    information on the CE's characteristics (e.g. SiteName;\n";
    print "                    in case they are not provided by the accounting record\n";
    print "                    itself).\n";
#    print "             \"glite\": not yet implemented.\n";
    print "             Default: none, only accounting information contained in the\n";
    print "             HLR database will be used.\n";
    print "-M           Keep a mapping from CE IDs to ExecutingSites as well as\n";
    print "             SpecInt2000 and SpecFloat2000 benchmarks up-to-date and use\n";
    print "             that mapping in case the site name cannot be retrieved in any\n";
    print "             other way (e.g. querying the GRIS for EDG resources).\n";
    print "-R           Retry to construct and insert LcgRecords that could not be\n";
    print "             successfully converted and inserted into the APEL database\n"; 
    print "             in previous executions of ".$prgname."\n";
    print "             (up to MAX_NUM_RETRIES). If a construction/insertion attempt\n";
    print "             is not successful, it will be tracked in ConversionErrors only\n";
    print "             if -R is specified.\n";
    print "-v           Verbose output: additional information of conversion of the\n";
    print "             single LcgRecord fields will be printed to stdout.\n";
    exit(1);

} # printUsageInfo() ...



# ---------------------------------------------------------------------------
#
# extract the sinle log entries from a result row (the current $row):
# argument: none
#
# ---------------------------------------------------------------------------
sub getTransInLogEntries {

    my @logTokens = split(/,/, $row->{'transLog'});
    my $token;
    foreach $token (@logTokens) {

	#print "[DGAS2APEL]              extracted from log: <$token>; ";

	my @entryParts = split(/=/, $token);

	#print "split into ".(scalar @entryParts)." parts.\n";
 
	if ($entryParts[0] ne "") {

	    # we found something meaningful:
	    my $entryName = $entryParts[0];
	    my $entryVal;

	    # check wether we splitted to many '=' (might be contained in the
	    # entry's value as well (such as cert subjects ...)
	    my $numEP = scalar(@entryParts);

	    if ($numEP == 2) {
		# exactly what we expect
		$entryVal = $entryParts[1];
	    }
	    elsif ($numEP > 2) {
		my $maxEPindex = $numEP - 1;
		# value was splitted because it contained '=', rebuild it:
		$entryVal = $entryParts[1]; #first part
		my $valPart;
		foreach $valPart (@entryParts[2 .. $maxEPindex]) {
		    $entryVal = $entryVal."=".$valPart; # additional parts
		}
	    }
	    else {
		# only one part, interpret it as a boolean value
		$entryVal = "";
	    }

	    # add the entry to the hash:
	    $logEntries{$entryName} = $entryVal;
	}
    }

    # transaction log already printed above, this was for testing if it is
    # parsed correctly ...
    #
    if ($verboseOutput) {
	my $logKey;
	foreach $logKey (keys(%logEntries)) {
	    print "[DGAS2APEL]              from transInLog.log: $logKey=<$logEntries{$logKey}>\n";
	}
    }

} # getTransInLogEntries() ...




# ---------------------------------------------------------------------------
#
# construct the APEL (LcgRecords) entry from the HLR logEntries:
# arguments: none
# returns: 1 if the APEL records contains all mandatory elements, otherwise 0.
#
# ---------------------------------------------------------------------------
sub constructAPELrecord() {

    # for the return value ...
    my $insertRecord = 1;

    # reset for this transaction.
    $GRISavailable = 1;

    my $globalCeId = "";

    if (exists($logEntries{"CE_ID"}) && ($logEntries{"CE_ID"} ne "")) {
	$globalCeId = $logEntries{"CE_ID"};
    }

    # constructing empty APEL record for the transaction:
    %lcgRecord = (
		  RecordIdentity => "NULL",
		  ExecutingSite => "NULL",
		  LocalJobID => "NULL",
		  LCGJobID => "NULL",
		  LocalUserID => "NULL",
		  LCGUserID => "NULL",
		  LCGUserVO => "NULL",
		  ElapsedTime => "NULL",
		  BaseCpuTime => "NULL",
		  ElapsedTimeSeconds => "NULL",
		  BaseCpuTimeSeconds => "NULL",
		  StartTime => "NULL",
		  StopTime => "NULL",
		  StartTimeUTC => "NULL",
		  StopTimeUTC => "NULL",
		  StartTimeEpoch => "NULL",
		  StopTimeEpoch => "NULL",
		  ExecutingCE => "NULL",
		  MemoryReal => "NULL",
		  MemoryVirtual => "NULL",
		  SpecInt2000 => "NULL",
		  SpecFloat2000 => "NULL",
		  EventDate => "0000-00-00",
		  EventTime => "00:00:00",
		  MeasurementDate => "0000-00-00",
		  MeasurementTime => "00:00:00" 
		  );

    #
    # setting lcgRecord fields:
    #
    if ($verboseOutput) {
	print "[DGAS2APEL]              -----\n[DGAS2APEL]              Building LcgRecords entry:\n";
    }

    # RecordIdentity:
    my $currDateTime = `date +%FT%T%z`;
    chomp($currDateTime);
    $lcgRecord{RecordIdentity} = "HLR=<$HLRLocation>; tid=<$row->{'transID'}>; conversion=<$currDateTime>";
    if ($verboseOutput) {
	print "[DGAS2APEL]              RecordIdentity=<$lcgRecord{RecordIdentity}>\n";
    }


    # ExecutingCE:
    # APEL takes only the hostname, not the entire CE ID.
    # (e.g. 'lcgce02.gridpp.rl.ac.uk')
    if (exists($logEntries{"CE_ID"}) && ($logEntries{"CE_ID"} ne "")) {
	my $ceHostname = ( split(/:/, $logEntries{"CE_ID"}) )[0];
	$lcgRecord{ExecutingCE} = $ceHostname;
    }
    if ($verboseOutput) {
	print "[DGAS2APEL]              ExecutingCE=<$lcgRecord{ExecutingCE}>\n";
    }
    if ($lcgRecord{ExecutingCE} eq "NULL") {
	print "[DGAS2APEL]              Error: Cannot determine ExecutingCE!\n";
	$insertRecord = 0;   # this was mandatory.
    }



    # ExecutingSite:
    my $foundSiteName = 0;

    if (exists($logEntries{"SiteName"}) && ($logEntries{"SiteName"} ne "")) {
	# SiteName in accounting log:
	$lcgRecord{ExecutingSite} = $logEntries{"SiteName"};
	$foundSiteName = 1;
    }
    else {
	# no SiteName in accounting log, try something else:

	if (exists($logEntries{"CE_ID"}) && ($logEntries{"CE_ID"} ne "")) {

	    # If the following methods don't work: ExecutingSite = hostname
	    my $ceHostname = ( split(/:/, $logEntries{"CE_ID"}) )[0];


	    # this is not yet the correct site name ... would only be used
	    # in case this is the last retry (forced insertion).
	    $lcgRecord{ExecutingSite} = $ceHostname;


	    # for EDG/LCG resources try to query the GRIS for the GlueSiteName
	    if ($queryGRIS) {

		my $glueSiteName =
		    &getValueFromGRISMultipleAttributes($ceHostname,
						     $NUM_GRIS_SITE_ATTRIBUTES,
						     \%GRIS_SITE_FILTERS,
						     \%GRIS_SITE_ATTRIBUTES);

		# if we're here we found something ...
		if ($glueSiteName ne ""
		    && $glueSiteName ne $GRIS_ERROR) {

		    $lcgRecord{ExecutingSite} = $glueSiteName;
		    $foundSiteName = 1;
		}
		else {
		    print "[DGAS2APEL]              Cannot get SiteName from GRIS!\n";
		}

	    } # if ($queryGRIS) ...

	    # for gLite resources: not yet implemented (here if dependend of
	    # CE_ID)

	} # if (exists($logEntries{"CE_ID"}) ...

	# for gLite resources: not yet implemented (here if independend of
	# CE_ID)

    } # no SiteName in accounting log (else)

    if (!$foundSiteName) {
	# alternative option: try to look up SiteName in MAP_CEID_ExecutingSite
	# if we don't have it already
	if ($useMapping) {
	    my $siteNameFromMap = &getLastSiteNameFromMap($globalCeId);
	    if ($siteNameFromMap ne "") {
		$foundSiteName = 1;
		print "[DGAS2APEL]              Cannot determine ExecutingSite! Taking last known site name.\n";
		$lcgRecord{ExecutingSite} = $siteNameFromMap;
	    }
	}
    }
    elsif ($insertRecord && $useMapping) {
	# record will most probably be added,
	# eventually update the mapping from CE ID to ExecutingSite
	&mapCEID2ExecutingSite($globalCeId, $lcgRecord{ExecutingSite});
    }

    if ($verboseOutput) {
	print "[DGAS2APEL]              ExecutingSite=<$lcgRecord{ExecutingSite}>\n";
    }

    # final check
    if (!$foundSiteName) {
	print "[DGAS2APEL]              Error: Cannot determine ExecutingSite!\n";
	$insertRecord = 0;   # this was mandatory.
    }


    # LocalJobID:
    if (exists($logEntries{"jobName"}) && ($logEntries{"jobName"} ne "")) {
	$lcgRecord{LocalJobID} = $logEntries{"jobName"};
    }
    if ($verboseOutput) {
	print "[DGAS2APEL]              LocalJobID=<$lcgRecord{LocalJobID}>\n";
    }

    # LCGJobID:
    $lcgRecord{LCGJobID} = $row->{'jobID'};
    if ($verboseOutput) {
	print "[DGAS2APEL]              LCGJobID=<$lcgRecord{LCGJobID}>\n";
    }

    # LocalUserID:
    if (exists($logEntries{"localUserId"}) && ($logEntries{"localUserId"} ne "")) {
	$lcgRecord{LocalUserID} = $logEntries{"localUserId"};
    }
    if ($verboseOutput) {
	print "[DGAS2APEL]              LocalUserID=<$lcgRecord{LocalUserID}>\n";
    }

    # LCGUserID: remains NULL,for reasons of privacy!
    # DON'T ADD THIS AS LONG AS APEL PUBLISHED THE DATA VIA R-GMA!
    if ($verboseOutput) {
	print "[DGAS2APEL]              LCGUserID=<$lcgRecord{LCGUserID}>\n";
    }

    # LCGUserVO:
    if (exists($logEntries{"userVo"}) && ($logEntries{"userVo"} ne "")) {
	$lcgRecord{LCGUserVO} = $logEntries{"userVo"};
    }
    if ($verboseOutput) {
	print "[DGAS2APEL]              LCGUserVO=<$lcgRecord{LCGUserVO}>\n";
    }

    # ElapsedTime and ElapsedTimeSeconds:
    if (exists($logEntries{"WALL_TIME"}) && ($logEntries{"WALL_TIME"} ne "")) {

	if ($logEntries{"WALL_TIME"} =~ /^\d*$/) { # make sure it's an integer
	    $lcgRecord{ElapsedTimeSeconds} = $logEntries{"WALL_TIME"};
	    $lcgRecord{ElapsedTime} = &timeSecs2Period($logEntries{"WALL_TIME"});
	}
    }
    if ($verboseOutput) {
	print "[DGAS2APEL]              ElapsedTime=<$lcgRecord{ElapsedTime}>\n";
	print "[DGAS2APEL]              ElapsedTimeSeconds=<$lcgRecord{ElapsedTimeSeconds}>\n";
    }

    # BaseCpuTime and BaseCpuTimeSeconds:
    if (exists($logEntries{"CPU_TIME"}) && ($logEntries{"CPU_TIME"} ne "")) {

	if ($logEntries{"CPU_TIME"} =~ /^\d*$/) { # make sure it's an integer
	    $lcgRecord{BaseCpuTimeSeconds} = $logEntries{"CPU_TIME"};
	    $lcgRecord{BaseCpuTime} = &timeSecs2Period($logEntries{"CPU_TIME"});
	}
    }
    if ($verboseOutput) {
	print "[DGAS2APEL]              BaseCpuTime=<$lcgRecord{BaseCpuTime}>\n";
	print "[DGAS2APEL]              BaseCpuTimeSeconds=<$lcgRecord{BaseCpuTimeSeconds}>\n";
    }
    if ($lcgRecord{BaseCpuTime} eq "NULL") {
	print "[DGAS2APEL]              Error: Cannot determine BaseCpuTime!\n";
	$insertRecord = 0;   # this was mandatory.
    }


    # StartTime, StartTimeEpoch and StartTimeUTC:
    # We leave StartTime NULL, if we don't have the timeZone. Use a mapping
    # table (ExecutingCE -> timeZone) optionally???
    if (exists($logEntries{"start"}) && ($logEntries{"start"} ne "")) {

	if ($logEntries{"start"} =~ /^\d*$/) { # make sure it's an integer
	    $lcgRecord{StartTimeEpoch} = $logEntries{"start"};
	    $lcgRecord{StartTimeUTC} = &timestamp2String($logEntries{"start"}, "Z");
	    if (exists($logEntries{"timeZone"}) && ($logEntries{"timeZone"} ne "")) {
		$lcgRecord{StartTime} = &timestamp2String($logEntries{"start"},$logEntries{"timeZone"});
	    }
	}
    }
    if ($verboseOutput) {
	print "[DGAS2APEL]              StartTime=<$lcgRecord{StartTime}>\n";
	print "[DGAS2APEL]              StartTimeUTC=<$lcgRecord{StartTimeUTC}>\n";
	print "[DGAS2APEL]              StartTimeEpoch=<$lcgRecord{StartTimeEpoch}>\n";
    }

    # StopTime, StopTimeEpoch and StopTimeUTC:
    # We leave StopTime NULL, if we don't have the timeZone. Use a mapping
    # table (ExecutingCE -> timeZone) optionally???
    if (exists($logEntries{"end"}) && ($logEntries{"end"} ne "")) {

	if ($logEntries{"end"} =~ /^\d*$/) { # make sure it's an integer
	    $lcgRecord{StopTimeEpoch} = $logEntries{"end"};
	    $lcgRecord{StopTimeUTC} = &timestamp2String($logEntries{"end"}, "Z");
	    if (exists($logEntries{"timeZone"}) && ($logEntries{"timeZone"} ne "")) {
		$lcgRecord{StopTime} = &timestamp2String($logEntries{"end"},$logEntries{"timeZone"});
	    }
	}
    }
    if ($verboseOutput) {
	print "[DGAS2APEL]              StopTime=<$lcgRecord{StopTime}>\n";
	print "[DGAS2APEL]              StopTimeUTC=<$lcgRecord{StopTimeUTC}>\n";
	print "[DGAS2APEL]              StopTimeEpoch=<$lcgRecord{StopTimeEpoch}>\n";
    }


    # MemoryReal:
    # APEL stores it as integer (in KBytes), not as a string:
    if (exists($logEntries{"MEM"}) && ($logEntries{"MEM"} ne "")) {
        if ($logEntries{"MEM"} =~ /^\s*(\d*)(.*)/) {
            # $1 is the amount, $2 (if it exists) the unit of measurement
            my $numKB = $1;
            if (exists($memoryUnitsGGF{$2})) {
		# APEL uses KBytes, what do we use?
		if ($verboseOutput) {
		    print "[DGAS2APEL]              memory units in log: $2 = $memoryUnitsGGF{$2} bits\n";
		    print "[DGAS2APEL]              memory units for APEL: KB = $memoryUnitAPEL bits\n";
		}
		my $factor = $memoryUnitsGGF{$2} / $memoryUnitAPEL;
		$numKB = int($numKB*$factor + 0.5);
            }
            else {
		if ($verboseOutput) {
		    print "[DGAS2APEL]              memory units not recognized in log, assuming KB.\n";
		}
            } 
            $lcgRecord{MemoryReal} = $numKB;
        }
    }
    if ($verboseOutput) {
	print "[DGAS2APEL]              MemoryReal=<$lcgRecord{MemoryReal}>\n";
    }

    # MemoryVirtual:
    # APEL stores it as integer (in KBytes), not as a string:
    if (exists($logEntries{"VMEM"}) && ($logEntries{"VMEM"} ne "")) {
        if ($logEntries{"VMEM"} =~ /^\s*(\d*)(.*)/) {
            # $1 is the amount, $2 (if it exists) the unit of measurement
            my $numKB = $1;
            if (exists($memoryUnitsGGF{$2})) {
                # APEL uses KBytes, what do we use?
		if ($verboseOutput) {
		    print "[DGAS2APEL]              memory units in log: $2 = $memoryUnitsGGF{$2} bits\n";
		    print "[DGAS2APEL]              memory units for APEL: KB = $memoryUnitAPEL bits\n";
		}
                my $factor = $memoryUnitsGGF{$2} / $memoryUnitAPEL;
                $numKB = int($numKB*$factor + 0.5);
            }
            else {
		if ($verboseOutput) {
		    print "[DGAS2APEL]              memory units not recognized in log, assuming KB.\n";
		}
            }
            $lcgRecord{MemoryVirtual} = $numKB;
        }
    }
    if ($verboseOutput) {
	print "[DGAS2APEL]              MemoryVirtual=<$lcgRecord{MemoryVirtual}>\n";
    }


    # SpecInt2000:
    my $foundSI00 = 0;

    if ($queryGRIS
	   && exists($logEntries{"CE_ID"}) && ($logEntries{"CE_ID"} ne "")) {
	# first try to get the value from the GRIS

	# If the following methods don't work: ExecutingSite = hostname
	my $ceHostname = ( split(/:/, $logEntries{"CE_ID"}) )[0];

	# try to query the GRIS:
	my $gridSI00 = 0;
	$gridSI00 = &getValueFromGRISMultipleAttributes($ceHostname,
						     $NUM_GRIS_SI00_ATTRIBUTES,
						     \%GRIS_SI00_FILTERS,
						     \%GRIS_SI00_ATTRIBUTES);
	if ($gridSI00 ne $GRIS_ERROR && $gridSI00 =~ /^\d*$/) {
	    $foundSI00 = 1; # found, even if 0 is returned from the GRIS

	    if ($gridSI00 != 0) {
		# but we leave NULL istead of 0 if not specified on GRIS.
		$lcgRecord{SpecInt2000} = $gridSI00;
	    }
	}
    } # if ($queryGRIS ...
    elsif (exists($logEntries{"specInt2000"})
	   && ($logEntries{"specInt2000"} ne "")) {
	# if not from GRIS, search in transInLog.log

	if ($logEntries{"specInt2000"} =~ /^\d*$/) { # make sure it's an int
	    $lcgRecord{SpecInt2000} = $logEntries{"specInt2000"};
	    $foundSI00 = 1;
	}
    }

    if (!$foundSI00 && $useMapping) {
	my $mapSI00 = &getLastSpecInt2000FromMap($globalCeId);
	if ($mapSI00 != 0) {
	    $lcgRecord{SpecInt2000} = $mapSI00;
	    print "[DGAS2APEL]              Cannot determine SpecInt2000! Taking last known value.\n";
	    # we don't set $foundSI00, because it we don't update the map if
	    # we took the value from there!
	    
	}
	else {
	    # don't insert the record, SpecInt is considered essential.
	    $insertRecord = 0;
	}
    }
    if ($verboseOutput) {
	print "[DGAS2APEL]              SpecInt2000=<$lcgRecord{SpecInt2000}>\n";
    }


    # SpecFloat2000:
    my $foundSF00 = 0;

    if ($queryGRIS
	   && exists($logEntries{"CE_ID"}) && ($logEntries{"CE_ID"} ne "")) {
	# first try to get the value from the GRIS

	# If the following methods don't work: ExecutingSite = hostname
	my $ceHostname = ( split(/:/, $logEntries{"CE_ID"}) )[0];

	# try to query the GRIS:
	my $gridSF00 = 0;
	$gridSF00 = &getValueFromGRISMultipleAttributes($ceHostname,
						     $NUM_GRIS_SF00_ATTRIBUTES,
						     \%GRIS_SF00_FILTERS,
						     \%GRIS_SF00_ATTRIBUTES);
	if ($gridSF00 ne $GRIS_ERROR && $gridSF00 =~ /^\d*$/) {
	    $foundSF00 = 1; # found, even if 0 is returned from the GRIS

	    if ($gridSF00 != 0) {
		# but we leave NULL istead of 0 if not specified on GRIS.
		$lcgRecord{SpecFloat2000} = $gridSF00;
	    }
	}
    }  # if ($queryGRIS ...
    elsif (exists($logEntries{"specFloat2000"})
	   && ($logEntries{"specFloat2000"} ne "")) {
	# if not from GRIS, search in transInLog.log

	if ($logEntries{"specFloat2000"} =~ /^\d*$/) { # make sure it's an int
	    $lcgRecord{SpecFloat2000} = $logEntries{"specFloat2000"};
	    $foundSF00 = 1;
	}
    }

    if (!$foundSF00 && $useMapping) {
	my $mapSF00 = &getLastSpecFloat2000FromMap($globalCeId);
	if ($mapSF00 != 0) {
	    $lcgRecord{SpecFloat2000} = $mapSF00;
	    print "[DGAS2APEL]              Cannot determine SpecFloat2000! Taking last known value.\n";
	    # we don't set $foundSF00, because it we don't update the map if
	    # we took the value from there!
	}
    }
    if ($verboseOutput) {
	print "[DGAS2APEL]              SpecFloat2000=<$lcgRecord{SpecFloat2000}>\n";
    }
    # eventually update map:
    if ($foundSI00 && $foundSF00 && $useMapping) {
	&mapCEID2Benchmarks($globalCeId, $lcgRecord{SpecInt2000},
			   $lcgRecord{SpecFloat2000});	
    }


    # MeasurementDate and MeasurementTime:
    # Extract it from the transaction timestamp.
    # For APEL this is the creation of the LcgRecords entry, for us it
    # is the creation of the HLR transaction! We however have the
    # creation time of the LcgRecord as a part of RecordIdentity
    # (conversion=<...>).
    my $transTimeUTC = &timestamp2String($row->{'transTimestamp'}, "Z");
    my @trTUTCtokens = split(/T|Z/, $transTimeUTC);
    if (exists($trTUTCtokens[0]) && $trTUTCtokens[0] ne ""
        && $trTUTCtokens[0] ne "NULL") {
        $lcgRecord{MeasurementDate} = $trTUTCtokens[0];
    }
    if (exists($trTUTCtokens[1]) && $trTUTCtokens[1] ne "") {
        $lcgRecord{MeasurementTime} = $trTUTCtokens[1];
    }
    # EventDate and EventTime:
    # For APEL this is the timestamp of the PBS/LSF event log. We assume it
    # to be equal to the stop time of the job. 
    # Extract the EventDate and Time from $lcgRecord{StopTimeUTC}.
    my @stpTUTCtokens = split(/T|Z/, $lcgRecord{StopTimeUTC});
    if (exists($stpTUTCtokens[0]) && $stpTUTCtokens[0] ne ""
        && $stpTUTCtokens[0] ne "NULL") {
        $lcgRecord{EventDate} = $stpTUTCtokens[0];
    }
    if (exists($stpTUTCtokens[1]) && $stpTUTCtokens[1] ne "") {
        $lcgRecord{EventTime} = $stpTUTCtokens[1];
    }
    # if EventDate and EventTime are not set then approximate them by
    # using the transaction timestamp (MeasurementDate/Time):
    if ($lcgRecord{EventDate} eq "0000-00-00" && $lcgRecord{EventTime} eq "00:00:00") {
        $lcgRecord{EventDate} = $lcgRecord{MeasurementDate};
        $lcgRecord{EventTime} = $lcgRecord{MeasurementTime};
    }
    if ($verboseOutput) {
	print "[DGAS2APEL]              EventDate=<$lcgRecord{EventDate}>\n";
	print "[DGAS2APEL]              EventTime=<$lcgRecord{EventTime}>\n";
	print "[DGAS2APEL]              MeasurementDate=<$lcgRecord{MeasurementDate}>\n";
	print "[DGAS2APEL]              MeasurementTime=<$lcgRecord{MeasurementTime}>\n";
    }

    return $insertRecord;

} # constructAPELrecord() ...



# ---------------------------------------------------------------------------
#
# insert an LcgRecords entry to the APEL database
# arguments: none
# returns: 1 if inserted, otherwise 0.
#
# ---------------------------------------------------------------------------
sub insertAPELrecord() {

    # Constructing insert statement for LcgRecord:
    $apelDbQuery = "INSERT INTO LcgRecords VALUES ("
        ."\"$lcgRecord{RecordIdentity}\","
        ."\"$lcgRecord{ExecutingSite}\","
        ."\"$lcgRecord{LocalJobID}\","
        ."\"$lcgRecord{LCGJobID}\","
        ."\"$lcgRecord{LocalUserID}\","
        ."\"$lcgRecord{LCGUserID}\","
        ."\"$lcgRecord{LCGUserVO}\","
        ."\"$lcgRecord{ElapsedTime}\","
        ."\"$lcgRecord{BaseCpuTime}\","
        ."$lcgRecord{ElapsedTimeSeconds},"
        ."$lcgRecord{BaseCpuTimeSeconds},"
        ."\"$lcgRecord{StartTime}\","
        ."\"$lcgRecord{StopTime}\","
        ."\"$lcgRecord{StartTimeUTC}\","
        ."\"$lcgRecord{StopTimeUTC}\","
        ."$lcgRecord{StartTimeEpoch},"
        ."$lcgRecord{StopTimeEpoch},"
        ."\"$lcgRecord{ExecutingCE}\","
        ."$lcgRecord{MemoryReal},"
        ."$lcgRecord{MemoryVirtual},"
        ."$lcgRecord{SpecInt2000},"
        ."$lcgRecord{SpecFloat2000},"
        ."\"$lcgRecord{EventDate}\","
        ."\"$lcgRecord{EventTime}\","
        ."\"$lcgRecord{MeasurementDate}\","
        ."\"$lcgRecord{MeasurementTime}\""
        .")";

   # check which values are NULL and do not specify them as a string
   # (that is, relpace >"NULL"< with >NULL<). 
   $apelDbQuery =~ s/"NULL"/NULL/g;

   # Insert the transaction into the APEL conversion db:

   print "[DGAS2APEL] Insert transaction into LcgRecords:\n[DGAS2APEL] Query = $apelDbQuery\n";

   eval { $apelSth = $apelDbh->prepare($apelDbQuery) };
   if ($@) {
       print STDERR "[DGAS2APEL] Query preparation failed! Error: $@\n";
       return 0; # could not insert
   }

   eval { $apelSth->execute() };
   if ($@) {
       if ($@ =~ /Duplicate entry/) {
           # not a real error, the record exists, try next
           print "[DGAS2APEL] Warning: LcgRecords entry for HLR tid=i$row->{'transID'} with RecordIdentity=<$lcgRecord{RecordIdentity}> ... skipping!\n";
       }
       else {
           # some other error, quit!
           print STDERR "[DGAS2APEL] Query execution failed! Error: $@\n";
           return 0; # could not insert
       }
   }

    return 1; # inserted (or already present; skipped)

} # insertAPELrecord() ...



# ---------------------------------------------------------------------------
#
# add an entry to the HLR tid -> LcgRecords RecordIdentity MAP
# arguments: none
# returns: 1 if inserted, otherwise 0.
#
# ---------------------------------------------------------------------------
sub addTid2RecordIdentityMAP() {

    $convDbQuery = "INSERT INTO MAP_tid_RecordIdentity VALUES ("
	."\"$HLRLocation\","
	."$row->{'transID'},"
	."\"$lcgRecord{RecordIdentity}\""
	.")";
    # check which values are NULL and do not specify them as a string
    # (that is, relpace >"NULL"< with >NULL<).
    $convDbQuery =~ s/"NULL"/NULL/g;

    # Insert the tid<->RecordIdentity mapping into the conversion log db:
    print "[DGAS2APEL] Insert tid<->RecordIdentity mapping into MAP_tid_RecordIdentity:\n[DGAS2APEL] Query = $convDbQuery\n"; 

    eval { $convSth = $convDbh->prepare($convDbQuery) };
    if ($@) {
	print STDERR "Query preparation failed! Error: $@\n";
	return 0; # map entry not inserted
    }

    eval { $convSth->execute() };
    if ($@) {
	if ($@ =~ /Duplicate entry/) {
	    # not a real error, the record exists, try next
	    print "[DGAS2APEL] Warning: tid<->RecordIdentity entry for HLR tid=i$row->{'transID'} and HLRLocation=<$HLRLocation> exists ... skipping!\n";
	}
	else {
	    # some other error, quit!
	    print STDERR "[DGAS2APEL] Query execution failed! Error: $@\n";
	    return 0; # not inserted.
	}
    }

    return 1; # inserted (or already present; skipped)

} # addTid2RecordIdentityMAP() ...




# ---------------------------------------------------------------------------
#
# add an entry to ConversionErrors or increment an existing entry (if a
# transaction could not be processed).
# argument: $tid (ID of the transaction that could not be processed)
# returns: 1 if inserted/incremented, otherwise 0.
#
# ---------------------------------------------------------------------------
sub incrementConversionErrorCount() {

    my $tid = $_[0];

    my $newEntry = 1;
    my $numRetries = 0;

    # determine current time:
    my $entryTime = &timestamp2String(time(), "Z");
    my $t_Pos = index($entryTime, "T");
    substr($entryTime, $t_Pos, 1) = " "; # overwrite T with a space
    my $z_Pos = index($entryTime, "Z");
    substr($entryTime, $z_Pos, 1) = "";  # delete trailing Z


    # first check if it already exists:
    $convDbQuery = "SELECT NumRetries FROM ConversionErrors WHERE "
	."HLRLocation=\"".$HLRLocation."\" AND "
	."HLR_tid=".$tid;

    eval { $convSth = $convDbh->prepare($convDbQuery) };
    if ($@) {
	print STDERR "Query preparation failed! Error: $@\n";
	return 0; # not inserted/incremented
    }

    eval { $convSth->execute() };
    if ($@) {
	# some other error, quit!
	print STDERR "[DGAS2APEL] Query execution failed! Error: $@\n";
	return 0; # not inserted/incremented
    }

    my $cErrRow = $convSth->fetchrow_hashref();
    if (exists($cErrRow->{'NumRetries'})) {
	$numRetries = int($cErrRow->{'NumRetries'});
	$newEntry = 0; # not new, increment
    }


    if ($newEntry) {

	# add a new entry to the table (setting NumRetries to 0):

	$convDbQuery = "INSERT INTO ConversionErrors VALUES ("
	    ."\"".$HLRLocation.     "\",".$tid.   ","
	    ."0,\"".$entryTime."\")";

	# Insert the error log into the db:
	if ($verboseOutput) {
	    print "[DGAS2APEL] Insert error log into ConversionErrors:\n[DGAS2APEL] Query = $convDbQuery\n"; 
	}

	eval { $convSth = $convDbh->prepare($convDbQuery) };
	if ($@) {
	    print STDERR "Query preparation failed! Error: $@\n";
	    return 0; # error entry not inserted
	}

	eval { $convSth->execute() };
	if ($@) {
	    if ($@ =~ /Duplicate entry/) {
		# not a real error, the record exists
		if ($verboseOutput) {
		    print "[DGAS2APEL] Warning: entry for HLRLocation="
			.$HLRLocation." and HLR_tid=".$tid
			." exists ... updating NumRetries!\n";
		}
		$newEntry = 0;
	    }
	    else {
		# some other error, quit!
		print STDERR "[DGAS2APEL] Query execution failed! Error: $@\n";
		return 0; # not inserted.
	    }
	}
    } # if ($newEntry)


    if (!$newEntry) {
	# update NumRetries and LastTryTimeUTC!

	$numRetries++;  # increment the number of retries previously determined

	if ($verboseOutput) {
	    print "[DGAS2APEL] Updating error entry in to NumRetries = ".$numRetries.".\n";
	}

	$convDbQuery = "UPDATE ConversionErrors SET "
	    ."LastTryTimeUTC=\"".$entryTime."\", "
	    ."NumRetries=".$numRetries
	    ." WHERE HLRLocation=\"".$HLRLocation."\""
	    ." AND HLR_tid=".$tid;

	eval { $convSth = $convDbh->prepare($convDbQuery) };
	if ($@) {
	    print STDERR "Query preparation failed! Error: $@\n";
	    return 0; # map entry not updated
	}

	eval { $convSth->execute() };
	if ($@) {
	    # some other error, quit!
	    print STDERR "[DGAS2APEL] Query execution failed! Error: $@\n";
	    return 0; # not updated.
	}

    } # if (!$newEntry) ...


    return 1; # inserted (or updated)

} # incrementConversionErrorCount() ...




# ---------------------------------------------------------------------------
#
# get the number of retries from an entry in ConversionErrors (if a 
# transaction could not be processed).
# argument: $tid (ID of the transaction that could not be processed)
# returns: NumRetries if the enrty was found, otherwise -1.
#
# ---------------------------------------------------------------------------
sub getNumRetriesForTid() {

    my $tid = $_[0];

    my $numRetries = -1;

    if ($verboseOutput) {
	print "[DGAS2APEL] Retrieving number of retries from ConversionErrors.\n";
    }    

    # first check if it already exists:
    $convDbQuery = "SELECT NumRetries FROM ConversionErrors WHERE "
	."HLRLocation=\"".$HLRLocation."\" AND "
	."HLR_tid=".$tid;

    eval { $convSth = $convDbh->prepare($convDbQuery) };
    if ($@) {
	print STDERR "Query preparation failed! Error: $@\n";
	return $numRetries; # not found
    }

    eval { $convSth->execute() };
    if ($@) {
	# some other error, quit!
	print STDERR "[DGAS2APEL] Query execution failed! Error: $@\n";
	return $numRetries; # not found
    }

    my $cErrRow = $convSth->fetchrow_hashref();
    if (exists($cErrRow->{'NumRetries'})) {
	$numRetries = int($cErrRow->{'NumRetries'});
    }

    if ($verboseOutput) {
	print "[DGAS2APEL] Number of retries: \n";
    }    

    return $numRetries; # NumRetries if found, otherwise -1

} # getNumRetriesForTid() ...




# ---------------------------------------------------------------------------
#
# add (or update) an entry to MAP_CEID_ExecutingSite for later
# retrieval (if a GRIS cannot be contacted for example).
# arguments: 1) CE_ID
#            2) ExecutingSite
# returns: 1 if inserted/updated (or already the same), otherwise 0.
#
# ---------------------------------------------------------------------------
sub mapCEID2ExecutingSite() {

    my $exCE = $_[0];
    my $exSite = $_[1];

    my $currSite;

    my $newEntry = 1;

    if ($exCE eq "") {
	if ($verboseOutput) {
	    print "[DGAS2APEL] Cannot insert mapping entry into MAP_CEID_ExecutingSite: no CE ID.\n";
	}
	return 0;  # not inserted/updated
    }

    # determine current time:
    my $entryTime = &timestamp2String(time(), "Z");
    my $t_Pos = index($entryTime, "T");
    substr($entryTime, $t_Pos, 1) = " "; # overwrite T with a space
    my $z_Pos = index($entryTime, "Z");
    substr($entryTime, $z_Pos, 1) = "";  # delete trailing Z


    # first check if it already exists:
    $convDbQuery = "SELECT ExecutingSite FROM MAP_CEID_ExecutingSite "
	."WHERE CEID=\"".$exCE."\"";

    eval { $convSth = $convDbh->prepare($convDbQuery) };
    if ($@) {
	print STDERR "Query preparation failed! Error: $@\n";
	return 0; # not inserted/updated
    }

    eval { $convSth->execute() };
    if ($@) {
	# some other error, quit!
	print STDERR "[DGAS2APEL] Query execution failed! Error: $@\n";
	return 0; # not inserted/updated
    }

    my $mapRow = $convSth->fetchrow_hashref();
    if (exists($mapRow->{'ExecutingSite'})) {
	$currSite = $mapRow->{'ExecutingSite'};
	$newEntry = 0; # not new, eventually update
    }

    if ($newEntry) {

	# add a new entry to the map:
	$convDbQuery = "INSERT INTO MAP_CEID_ExecutingSite VALUES ("
	    ."\"".$exCE.     "\",\"".$exSite.   "\","
	    ."\"".$entryTime."\",\"".$entryTime."\")";

	# Insert the CE->Site mapping into the db:
	if ($verboseOutput) {
	    print "[DGAS2APEL] Insert mapping entry into MAP_CEID_ExecutingSite:\n[DGAS2APEL] Query = $convDbQuery\n";
	}

	eval { $convSth = $convDbh->prepare($convDbQuery) };
	if ($@) {
	    print STDERR "Query preparation failed! Error: $@\n";
	    return 0; # map entry not inserted
	}

	eval { $convSth->execute() };
	if ($@) {
	    if ($@ =~ /Duplicate entry/) {
		# not a real error, the record exists
		if ($verboseOutput) {
		    print "[DGAS2APEL] Warning: entry for CEID=".$exCE." exists ... updating!\n";
		}
		$newEntry = 0;
	    }
	    else {
		# some other error, quit!
		print STDERR "[DGAS2APEL] Query execution failed! Error: $@\n";
		return 0; # not inserted.
	    }
	}
    } # if ($newEntry)

    if (!$newEntry) {
	# update at least ValidityTimeUTC!

	$convDbQuery = "UPDATE MAP_CEID_ExecutingSite SET "
	    ."ValidityTimeUTC=\"".$entryTime."\" ";

	if ($currSite eq $exSite) {
	    if ($verboseOutput) {
		print "[DGAS2APEL] Mapping entry in MAP_CEID_ExecutingSite up to date\n";
		print "[DGAS2APEL] (updating only ValidityTimeUTC to current date & time).\n";
	    }
	}
	else {
	    # old entry is different, update also ExecutingSite:
	    if ($verboseOutput) {
		print "[DGAS2APEL] Updating mapping entry in MAP_CEID_ExecutingSite to ".$exSite.".\n";
	    }

	    $convDbQuery = $convDbQuery.", ExecutingSite=\"".$exSite."\" "
		.", UpdateTimeUTC=\"".$entryTime."\" ";
	}


	$convDbQuery = $convDbQuery."WHERE CEID=\"".$exCE."\"";

	eval { $convSth = $convDbh->prepare($convDbQuery) };
	if ($@) {
	    print STDERR "Query preparation failed! Error: $@\n";
	    return 0; # map entry not updated
	}

	eval { $convSth->execute() };
	if ($@) {
	    # some other error, quit!
	    print STDERR "[DGAS2APEL] Query execution failed! Error: $@\n";
	    return 0; # not updated.
	}

    } # if (!$newEntry) ...

    return 1; # inserted (or updated if it has changed)

} # mapCEID2ExecutingSite() ...




# ---------------------------------------------------------------------------
#
# retrieve an entry to MAP_CEID_ExecutingSite
# (if a GRIS cannot be contacted for example).
# argument: CE_ID
# returns: site name if found, otherwise "".
#
# ---------------------------------------------------------------------------
sub getLastSiteNameFromMap() {

    my $exCE = $_[0];

    my $exSiteName = "";

    if ($exCE eq "") {
	if ($verboseOutput) {
	    print "[DGAS2APEL] Cannot get site name from MAP_CEID_ExecutingSite: no CE ID.\n";
	}
	return $exSiteName;  # undefined, not found
    }

    # check if it exists:
    $convDbQuery = "SELECT ExecutingSite FROM MAP_CEID_ExecutingSite "
	."WHERE CEID=\"".$exCE."\"";

    eval { $convSth = $convDbh->prepare($convDbQuery) };
    if ($@) {
	print STDERR "Query preparation failed! Error: $@\n";
	return $exSiteName;  # undefined, not found
    }

    eval { $convSth->execute() };
    if ($@) {
	# some other error, quit!
	print STDERR "[DGAS2APEL] Query execution failed! Error: $@\n";
	return $exSiteName; # undefined, not found
    }

    my $mapRow = $convSth->fetchrow_hashref();
    if (exists($mapRow->{'ExecutingSite'})) {
	$exSiteName = $mapRow->{'ExecutingSite'};
    }

    if ($exSiteName eq "" && $verboseOutput) {
	print "[DGAS2APEL] Cannot get site name from MAP_CEID_ExecutingSite: no entry.\n";
    }

    return $exSiteName; # the site name, if found, otherwise undefined.

} # getLastSiteNameFromMap() ...



# ---------------------------------------------------------------------------
#
# retrieve tids and numRetries from ConversionErrors.
# arguments: none
# returns: %tids2NumRetries (tids as keys, number of retries as values).
#
# ---------------------------------------------------------------------------
sub getTransactionsWithConversionErrors() {

    my %tids2NR = ();

    # check if it exists:
    $convDbQuery = "SELECT HLR_tid,NumRetries FROM ConversionErrors "
	."WHERE HLRLocation=\"".$HLRLocation."\"";

    eval { $convSth = $convDbh->prepare($convDbQuery) };
    if ($@) {
	print STDERR "Query preparation failed! Error: $@\n";
	return %tids2NR;  # undefined, not found
    }

    eval { $convSth->execute() };
    if ($@) {
	# some other error, quit!
	print STDERR "[DGAS2APEL] Query execution failed! Error: $@\n";
	return %tids2NR; # undefined, not found
    }


    my $convErrRow;
    while ($convErrRow = $convSth->fetchrow_hashref()) {
	if (exists($convErrRow->{'HLR_tid'})
	    && exists($convErrRow->{'NumRetries'}) ) {

	    # add the tid->numRetries mapping:
	    $tids2NR{$convErrRow->{'HLR_tid'}} = int($convErrRow->{'NumRetries'});
	}
    }

    return %tids2NR; # tids->numRetries map, if found, otherwise undefined.

} # getTransactionsWithConversionErrors() ...



# ---------------------------------------------------------------------------
#
# remove a specific entry from ConversionErrors.
# arguments: tid (transaction id).
# returns: 1 if removed, otherwise 0.
#
# ---------------------------------------------------------------------------
sub removeConversionError() {

    my $rmTid = $_[0];

    if ($rmTid eq "") {
	if ($verboseOutput) {
	    print "[DGAS2APEL] Cannot remove entry from ConversionErrors: no HLR_tid.\n";
	}
	return 0;  # not removed
    }

    # check if it exists:
    $convDbQuery = "DELETE FROM ConversionErrors "
	."WHERE HLRLocation=\"".$HLRLocation."\" AND HLR_tid=".$rmTid;

    eval { $convSth = $convDbh->prepare($convDbQuery) };
    if ($@) {
	print STDERR "Query preparation failed! Error: $@\n";
	return 0;  # not removed
    }

    eval { $convSth->execute() };
    if ($@) {
	# some other error, quit!
	print STDERR "[DGAS2APEL] Query execution failed! Error: $@\n";
	return 0;  # not removed
    }

    return 1; # removed.

} # removeConversionError() ...




# ---------------------------------------------------------------------------
#
# query a GRIS an retrieve a specific value.
# arguments: $grisHost, $grisPort, $grisFilter, $searchAttribute
# returns: the value associated to the Attribute, otherwise $GRIS_ERROR;
#
# ---------------------------------------------------------------------------
sub getValueFromGRIS() {

    my $grisHost = $_[0];
    my $grisPort = $_[1];
    my $grisFilter = $_[2];
    my $searchAttribute = $_[3];

    my $retVal = $GRIS_ERROR; # if nothing is found

    my $ldap;
    my $msg;

    # this is to make sure the attempt to contact a GRIS is made only once
    # per transaction, otherwise, if a GRIS temporarily is not available, the
    # execution will take far longer due to the frequent GRIS timeouts.
    if (!$GRISavailable) {
	print STDERR "[DGAS2APEL]              Cannot contact the GRIS.\n";
	return $GRIS_ERROR; # error
    }


    eval { # make sure the program doesn't quit if the LDAP
	   # server cannot be contacted.
	$ldap = Net::LDAP->new($grisHost, port=>$grisPort,
				  timeout=>$GRIS_TIMEOUT);
	$msg = $ldap->bind;
    };
    if ($@) {
	if ($verboseOutput) {
	    print STDERR "[DGAS2APEL]              Cannot contact the GRIS.\n";
	}
	$GRISavailable = 0;
	return $GRIS_ERROR; # error
    }


    eval { # make sure the program doesn't quit if the value cannot be retrieved
	if ($verboseOutput) {
	    print "[DGAS2APEL]              Trying to get '$searchAttribute' from GRIS with filter '$grisFilter'.\n";
	}
	$msg = $ldap->search(filter=>$grisFilter, base=>$GRIS_LDAPBASE);
	my @entries = $msg->entries;

	# split attribute, it might be something like
	# "GlueForeignKey: GlueSiteUniqueID=INFN-T1"
	my @attributeTokens = split(/: /, $searchAttribute);
	if (scalar(@attributeTokens) == 1) {
	    # also accept "GlueForeignKey:GlueSiteUniqueID=INFN-T1" (no space)
	    @attributeTokens = split(/:/, $searchAttribute);
	}

	# only one entry = object considered!
	my @attrValues = $entries[0]->get_value($attributeTokens[0]);

	my $attVal;
	foreach $attVal (@attrValues) { 

	    if (scalar(@attributeTokens) == 1 || $attributeTokens[1] eq "") {
		# no further specification, we take the first one:
		$retVal = $attVal;
		last;
	    }
	    else {
		# something like "GlueForeignKey: GlueSiteUniqueID=INFN-T1"
		# search all attribute/value pairs for the first 
		# occurrence of the extension:
		if ($attVal =~ /$attributeTokens[1]=(.*)/) {
		    $retVal = $1;
		    last;
		}
	    }
	}
    };
    if ($@) {
	if ($verboseOutput) {
	    print STDERR "[DGAS2APEL]              Cannot retrieve the value from the GRIS.\n";
	}
	return $GRIS_ERROR; # error
    }

    return $retVal;

} # getValueFromGRIS() ...




# ---------------------------------------------------------------------------
#
# add (or update) an entry to MAP_CEID_Benchmarks for later
# retrieval (if a GRIS cannot be contacted for example).
# arguments: 1) CE_ID
#            2) SpecInt2000 (can also be "NULL")
#            3) SpecFloat2000 (can also be "NULL")
# returns: 1 if inserted/updated (or already the same), otherwise 0.
#
# ---------------------------------------------------------------------------
sub mapCEID2Benchmarks() {

    my $exCE = $_[0];
    my $si2k = $_[1];
    my $sf2k = $_[2];

    my $currSi2k;
    my $currSf2k;

    my $newEntry = 1;

    if ($exCE eq "") {
	if ($verboseOutput) {
	    print "[DGAS2APEL] Cannot insert mapping entry into MAP_CEID_Benchmarks: no CE ID.\n";
	}
	return 0;  # not inserted/updated
    }

    # determine current time:
    my $entryTime = &timestamp2String(time(), "Z");
    my $t_Pos = index($entryTime, "T");
    substr($entryTime, $t_Pos, 1) = " "; # overwrite T with a space
    my $z_Pos = index($entryTime, "Z");
    substr($entryTime, $z_Pos, 1) = "";  # delete trailing Z


    # first check if it already exists:
    $convDbQuery = "SELECT SpecInt2000,SpecFloat2000 FROM MAP_CEID_Benchmarks "
	."WHERE CEID=\"".$exCE."\"";

    eval { $convSth = $convDbh->prepare($convDbQuery) };
    if ($@) {
	print STDERR "Query preparation failed! Error: $@\n";
	return 0; # not inserted/updated
    }

    eval { $convSth->execute() };
    if ($@) {
	# some other error, quit!
	print STDERR "[DGAS2APEL] Query execution failed! Error: $@\n";
	return 0; # not inserted/updated
    }

    my $mapRow = $convSth->fetchrow_hashref();
    if (exists($mapRow->{'SpecInt2000'})) {
	$currSi2k = $mapRow->{'SpecInt2000'};
	$newEntry = 0; # not new, eventually update
    }
    if (exists($mapRow->{'SpecFloat2000'})) {
	$currSf2k = $mapRow->{'SpecFloat2000'};
	$newEntry = 0; # not new, eventually update
    }

    if ($newEntry) {

	# add a new entry to the map:
	$convDbQuery = "INSERT INTO MAP_CEID_Benchmarks VALUES ("
	    ."\"".$exCE.     "\",".$si2k.   ",".$sf2k.   ","
	    ."\"".$entryTime."\",\"".$entryTime."\")";

	# Insert the CE->Site mapping into the db:
	if ($verboseOutput) {
	    print "[DGAS2APEL] Insert mapping entry into MAP_CEID_Benchmarks:\n[DGAS2APEL] Query = $convDbQuery\n";
	}

	eval { $convSth = $convDbh->prepare($convDbQuery) };
	if ($@) {
	    print STDERR "Query preparation failed! Error: $@\n";
	    return 0; # map entry not inserted
	}

	eval { $convSth->execute() };
	if ($@) {
	    if ($@ =~ /Duplicate entry/) {
		# not a real error, the record exists
		if ($verboseOutput) {
		    print "[DGAS2APEL] Warning: entry for CEID=".$exCE." exists ... updating!\n";
		}
		$newEntry = 0;
	    }
	    else {
		# some other error, quit!
		print STDERR "[DGAS2APEL] Query execution failed! Error: $@\n";
		return 0; # not inserted.
	    }
	}
    } # if ($newEntry)

    if (!$newEntry) {
	# update at least ValidityTimeUTC!

	$convDbQuery = "UPDATE MAP_CEID_Benchmarks SET "
	    ."ValidityTimeUTC=\"".$entryTime."\" ";

	my $updating = 0;

	# SpecInt2000
	if (defined($si2k ) && $si2k ne "NULL") {
	    if (defined($currSi2k) && $currSi2k eq $si2k) {
		if ($verboseOutput) {
		    print "[DGAS2APEL] SpecInt2000 in MAP_CEID_Benchmarks up to date\n";
		    print "[DGAS2APEL] (updating only ValidityTimeUTC to current date & time).\n";
		}
	    }
	    else {
		# old entry is different, update also ExecutingSite:
		if ($verboseOutput) {
		    print "[DGAS2APEL] Updating SpecInt2000 in MAP_CEID_Benchmarks to ".$si2k.".\n";
		}

		$convDbQuery = $convDbQuery
		    .", SpecInt2000=".$si2k." ";
		$updating = 1;
	    }
	}

	# SpecFloat2000
	if (defined($sf2k) && $sf2k ne "NULL") {
	    if (defined($currSf2k) && $currSf2k eq $sf2k) {
		if ($verboseOutput) {
		    print "[DGAS2APEL] SpecFloat2000 in MAP_CEID_Benchmarks up to date\n";
		    print "[DGAS2APEL] (updating only ValidityTimeUTC to current date & time).\n";
		}
	    }
	    else {
		# old entry is different, update also ExecutingSite:
		if ($verboseOutput) {
		    print "[DGAS2APEL] Updating SpecFloat2000 in MAP_CEID_Benchmarks to ".$sf2k.".\n";
		}

		$convDbQuery = $convDbQuery
		    .", SpecFloat2000=".$sf2k." ";
		$updating = 1;
	    }
	}

	if ($updating) {
	    $convDbQuery = $convDbQuery.", UpdateTimeUTC=\"".$entryTime."\" ";
	}

	$convDbQuery = $convDbQuery."WHERE CEID=\"".$exCE."\"";

	eval { $convSth = $convDbh->prepare($convDbQuery) };
	if ($@) {
	    print STDERR "Query preparation failed! Error: $@\n";
	    return 0; # map entry not updated
	}

	eval { $convSth->execute() };
	if ($@) {
	    # some other error, quit!
	    print STDERR "[DGAS2APEL] Query execution failed! Error: $@\n";
	    return 0; # not updated.
	}

    } # if (!$newEntry) ...

    return 1; # inserted (or updated if it has changed)

} # mapCEID2Benchmarks() ...




# ---------------------------------------------------------------------------
#
# retrieve an entry to MAP_CEID_Benchmarks
# (if a GRIS cannot be contacted for example).
# argument: CE_ID
# returns: SpecInt2000 if found, otherwise 0.
#
# ---------------------------------------------------------------------------
sub getLastSpecInt2000FromMap() {

    my $exCE = $_[0];

    my $exSI00 = -1;

    if ($exCE eq "") {
	if ($verboseOutput) {
	    print "[DGAS2APEL] Cannot get SpecInt2000 from MAP_CEID_Benchmarks: no CE ID.\n";
	}
	return 0;  # undefined, not found
    }

    # check if it exists:
    $convDbQuery = "SELECT SpecInt2000,SpecFloat2000 FROM MAP_CEID_Benchmarks "
	."WHERE CEID=\"".$exCE."\"";

    eval { $convSth = $convDbh->prepare($convDbQuery) };
    if ($@) {
	print STDERR "Query preparation failed! Error: $@\n";
	return 0;  # undefined, not found
    }

    eval { $convSth->execute() };
    if ($@) {
	# some other error, quit!
	print STDERR "[DGAS2APEL] Query execution failed! Error: $@\n";
	return 0; # undefined, not found
    }

    my $mapRow = $convSth->fetchrow_hashref();

    if (exists($mapRow->{'SpecInt2000'})
	&& defined($mapRow->{'SpecInt2000'})) {
	# if it is not NULL in the db:
	$exSI00 = int($mapRow->{'SpecInt2000'});
    }

    if ($exSI00 == -1) {
	# this won't happen if the value returned by the GRIS is 0 (it isn't
	# always set)

	$exSI00 = 0;

	if ($verboseOutput) {
	    print "[DGAS2APEL] Cannot get SpecInt2000 from MAP_CEID_Benchmarks: no entry.\n";
	}
    }

    return $exSI00; # SpecInt2000, if found, otherwise 0.

} # getLastSpecInt2000FromMap() ...



# ---------------------------------------------------------------------------
#
# retrieve an entry to MAP_CEID_Benchmarks
# (if a GRIS cannot be contacted for example).
# argument: CE_ID
# returns: SpecFloat2000 if found, otherwise 0.
#
# ---------------------------------------------------------------------------
sub getLastSpecFloat2000FromMap() {

    my $exCE = $_[0];

    my $exSF00 = -1;

    if ($exCE eq "") {
	if ($verboseOutput) {
	    print "[DGAS2APEL] Cannot get SpecFloat2000 from MAP_CEID_Benchmarks: no CE ID.\n";
	}
	return 0;  # undefined, not found
    }

    # check if it exists:
    $convDbQuery = "SELECT SpecInt2000,SpecFloat2000 FROM MAP_CEID_Benchmarks "
	."WHERE CEID=\"".$exCE."\"";

    eval { $convSth = $convDbh->prepare($convDbQuery) };
    if ($@) {
	print STDERR "Query preparation failed! Error: $@\n";
	return 0;  # undefined, not found
    }

    eval { $convSth->execute() };
    if ($@) {
	# some other error, quit!
	print STDERR "[DGAS2APEL] Query execution failed! Error: $@\n";
	return 0; # undefined, not found
    }

    my $mapRow = $convSth->fetchrow_hashref();

    if (exists($mapRow->{'SpecFloat2000'})
	&& defined($mapRow->{'SpecFloat2000'})) {
	# if it is not NULL in the db:
	$exSF00 = int($mapRow->{'SpecFloat2000'});
    }

    if ($exSF00 == -1) {
	# this won't happen if the value returned by the GRIS is 0 (it isn't
	# always set)

	$exSF00 = 0;

	if ($verboseOutput) {
	    print "[DGAS2APEL] Cannot get SpecFloat2000 from MAP_CEID_Benchmarks: no entry.\n";
	}
    }

    return $exSF00; # SpecFloat2000, if found, otherwise undefined.

} # getLastSpecFloat2000FromMap() ...



# ---------------------------------------------------------------------------
#
# query a GRIS for a certain value that may be associated to different
# attributes.
# (if a GRIS cannot be contacted for example).
# arguments: $ceHostname,
#            $NUM_GRIS_..._ATTRIBUTES,
#            \%GRIS_..._FILTERS,          <--- reference!
#            \%GRIS_..._ATTRIBUTES        <--- reference!
# returns: the value if found, otherwise $GRIS_ERROR.
#
# ---------------------------------------------------------------------------
sub getValueFromGRISMultipleAttributes() {

    my $grisHost = $_[0];
    my $numAtt = $_[1];
    my %grisFilters = %{$_[2]};     # dereferencing
    my %grisAttributes = %{$_[3]};  # dereferencing

    my $grisFilterCounter;

    my $retValue = $GRIS_ERROR;
    my $foundValue = 0;

    for ($grisFilterCounter = 1; $grisFilterCounter <= $numAtt;
	 $grisFilterCounter++) {

	if ($foundValue == 0) { # not yet found, try next filter/attribute pair

	    my $hashKey = "type".$grisFilterCounter;

	    $retValue = &getValueFromGRIS($grisHost, $GRIS_PORT,
					  $grisFilters{$hashKey},
					  $grisAttributes{$hashKey});

	    # if we're here we found something ...
	    if ($retValue ne $GRIS_ERROR && $retValue ne "") {
		$foundValue = 1;
	    }

	} # if ($foundValue == 0) ...
    } # for ...

    if (!$foundValue) {
	return $GRIS_ERROR;
    }

    return $retValue;

} # getValueFromGRISMultipleAttributes() ...
