#!/usr/bin/perl -w

use strict;

use Getopt::Std;

use DBI();

# command buffering of STDOUT
$| = 1;


# pre-defined return values
my $QUERY_ERROR = "DB_QUERY_ERROR";
my $QUERY_OK = "DB_QUERY_OK";
my $QUERY_NOT_FOUND = "DB_QUERY_NOT_FOUND";

my $prgname = "glite-dgas-hlr_updateDB.pl";

if ( !exists($ENV{GLITE_LOCATION}) || $ENV{GLITE_LOCATION} eq "" ) {
        $ENV{GLITE_LOCATION}="/opt/glite/";
	print "[updateDB] Environment variable GLITE_LOCATION not defined! Trying /opt/glite/ ...\n";
}

my $cfname = "dgas_hlr.conf";
my $conffile = "$ENV{GLITE_LOCATION}/etc/$cfname";

# -------------------
# getting options ...
# -------------------

my %opts;
$opts{h} = "";
$opts{c} = "";
$opts{v} = "";
$opts{r} = "";

my $useMysqlRootUser = 0;
my $rootPassword = "";

my $verboseOutput = 0;


getopts('hc:r:v', \%opts);

if ( $opts{h} eq "1" ) {
    &printUsageInfo();
}

if ( $opts{c} ne "" ) {
    $conffile = $opts{c};
}

if ( $opts{r} ne "" ) {
    $rootPassword = $opts{r};
    $useMysqlRootUser = 1;
}

if ( $opts{v} eq "1" ) {
    $verboseOutput = 1;
}


# ------------------------------
# parsing of configuration file:
# ------------------------------

# defaults for DGAS HLR db:
# This is the database where the original HLR accounting information is taken
# from.
my $hlrDbServer = "localhost";
my $hlrDbPort = 3306;
my $hlrDbUser = "root";
my $hlrDbPassword = "";
my $hlrDbName = "hlr";
my $hlrTmpDbName = "hlr_tmp";

# read configuration from file:
&readConfiguration();

if ($useMysqlRootUser) { # if the MySQL root password was specified on the
                         # command line, overwrite what we have taken from
                         # the HLR configuration file.
    $hlrDbUser = "root";
    $hlrDbPassword = $rootPassword;
    if ($verboseOutput) {
	print "[updateDB] Root password specified, using MySQL root user.\n";
    }
}
else {
    # don't use the root user, but the user specified in the config. file
    if ($verboseOutput) {
	print "[updateDB] Root password not specified, trying to use HLR MySQL user.\n";
    }
}


# -----------------------
# connecting to MySQL dbs
# -----------------------

my $hlrDbh;
my $hlrDbQuery;
my $hlrSth;

my $hlrTmpDbh;
my $hlrTmpDbQuery;
my $hlrTmpSth;

my $row;

&connectToDatabases();


# ------------------------------------
# processing the single DB updates ...
# ------------------------------------

my $exitStatus = 0; # 0 if ok

$exitStatus += &updateHlrTmp_trans_queue();   # add error status.

$exitStatus += &updateHlr_acctdesc();   # add error status.

$exitStatus += &updateHlr_transInInfo();   # add error status.

$exitStatus += &updateHlr_transOutInfo();   # add error status.

$exitStatus += &updateHlr_transInLog();   # add error status.

$exitStatus += &updateHlr_transOutLog();   # add error status.

$exitStatus += &updateHlr_jobAuth();   # add error status.

$exitStatus += &updateHlr_hlrAdmin();   # add error status.

$exitStatus += &updateHlr_voAdmin();   # add error status.

$exitStatus += &updateHlr_groupAdmin();   # add error status.


# ----------------------------------------
# finished updating procedure, exiting ...
# ----------------------------------------

exit ($exitStatus);



#############################################################################
# ===========================================================================
#                               FUNCTIONS
# ===========================================================================
#############################################################################



# ---------------------------------------------------------------------------
#
# Print usage information (help).
# arguments: none
#
# ---------------------------------------------------------------------------
sub printUsageInfo() {
    print "Author: Rosario Piro (piro\@to.infn.it)\n\n";
    print "Usage:\n";
    print "$prgname [OPTIONS]\n";
    print "\nWhere OPTIONS are:\n";
    print "-c <file>    Specifies the configuration file.\n";
    print "             Default: \$GLITE_LOCATION/etc/$cfname\n";
    print "-r <passwd>  Specifies the MySQL root password. If not specified on the\n";
    print "             command line, the tool tries to use the HLR MySQL user/password\n";
    print "             specified in the HLR configuration file. This will work only, if\n";
    print "             HLR MySQL user has the privileges to modify the HLR databases.\n";
    print "-v           Verbose output: more detailed information written to stdout.\n";
    exit(1);

} # printUsageInfo() ...





# ---------------------------------------------------------------------------
#
# read configuration file and parse parameters:
# arguments: none
#
# ---------------------------------------------------------------------------
sub readConfiguration() {

    # first read the file:
    my $line = "";

    unless ( -e $conffile ) {
	print STDERR "[updateDB] Error: configuration file $conffile not found!\n";
	exit(1);
    }

    unless ( -r $conffile ) {
	print STDERR "[updateDB] Error: configuration file $conffile not readable!\n";
	exit(1);
    }
         
    unless ( open(IN, "< $conffile") ) {
	print STDERR "[updateDB] Error: cannot open configuration file $conffile!\n";
        exit(1);
    }

    my %confparameters;

    # reading the single lines
    while ($line = <IN>) {
	chomp($line);

	$line || next;

	next if ($line =~ /^\s*\#+/);

	if ($line =~ /(\w*)\s*=\s*\"(.*)\"/) {

	    $confparameters{$1} = $2;

	}

    }

    #my $confpar;
    #foreach $confpar (keys(%confparameters)) {
    #    print "$confpar = $confparameters{$confpar}\n";
    #}

    # close the configuration file
    close(IN);


    # now parse the parameters:

    # get DGAS HLR db info:

    if (exists($confparameters{"hlr_sql_server"})) {
	$hlrDbServer = $confparameters{"hlr_sql_server"};
    }
    else {
	print "[updateDB] Configuration parameter hlr_sql_server missing, trying '$hlrDbServer'!\n";
    }

    if (exists($confparameters{"hlr_sql_user"})) {
	$hlrDbUser = $confparameters{"hlr_sql_user"};
    }
    else {
	print "[updateDB] Configuration parameter hlr_sql_user missing, trying '$hlrDbUser'!\n";
    }

    if (exists($confparameters{"hlr_sql_password"})) {
	$hlrDbPassword = $confparameters{"hlr_sql_password"};
    }
    else {
	print "[updateDB] Configuration parameter hlr_sql_password missing, trying '$hlrDbPassword'!\n";
    }

    if (exists($confparameters{"hlr_sql_dbname"})) {
	$hlrDbName = $confparameters{"hlr_sql_dbname"};
    }
    else {
	print "[updateDB] Configuration parameter hlr_sql_dbname missing, trying '$hlrDbName'!\n";
    }

    if (exists($confparameters{"hlr_tmp_sql_dbname"})) {
	$hlrTmpDbName = $confparameters{"hlr_tmp_sql_dbname"};
    }
    else {
	print "[updateDB] Configuration parameter hlr_tmp_sql_dbname missing, trying '$hlrTmpDbName'!\n";
    }

} # readConfiguration() ...



# ---------------------------------------------------------------------------
#
# connecting to the HLR databases
# argument: none
#
# ---------------------------------------------------------------------------
sub connectToDatabases() {

    # DGAS HLR db:

    eval { $hlrDbh = DBI->connect("DBI:mysql:database=$hlrDbName;host=$hlrDbServer",
				  "$hlrDbUser", "$hlrDbPassword",
				  {'RaiseError' => 1}
				  )
	   };
    if ($@) {
	print STDERR "[updateDB] Error: Cannot connect to DGAS HLR database! Error: $@\n";
	exit(1);
    }

    # DGAS HLR_TMP db:

    eval { $hlrTmpDbh = DBI->connect("DBI:mysql:database=$hlrTmpDbName;host=$hlrDbServer",
				  "$hlrDbUser", "$hlrDbPassword",
				  {'RaiseError' => 1}
				  )
	   };
    if ($@) {
	print STDERR "[updateDB] Error: Cannot connect to DGAS HLR TMP database! Error: $@\n";
	exit(1);
    }

} # connectToDatabases() ...



# ---------------------------------------------------------------------------
#
# arguments: timestamp, timeZone(e.g. "+0200")
# returns a local time if the second argument is not "+0000" or "Z"
#
# ---------------------------------------------------------------------------
sub timestamp2String() {

    my $tstamp = $_[0];
    my $tzStr = $_[1];

    # parse timeZone:
    if ($tzStr eq "+0000" || $tzStr eq "Z") {
	$tzStr = "Z";
    }
    elsif ($tzStr =~ /^(\+|-)\d{4}/) {
	my $hourDiff;
	my $minDiff;

	$hourDiff = int(substr($tzStr,0,3));
	$minDiff = int(substr($tzStr,3,2));

#	if ($hourDiff < 0) { <- this is wrong since it misses e.g. -0030 !
	if ($tzStr =~ /^-/) {
	    # negative time zone, consider minutes negative as well!
	    $minDiff = $minDiff * (-1);
	}

	# compute difference to UTC:
	$tstamp = $tstamp + ($hourDiff * 3600) + ($minDiff * 60);
    }
    else {
	if ($verboseOutput) {
	    print "[updateDB]              Wrong timezone format: $tzStr - cannot convert timestamp\n";
	}
	return "NULL";
    }

    my $sTime = "";

    eval {
	$sTime = gmtime $tstamp;
    };
    if ($@) {
	if ($verboseOutput) {
	    print "[updateDB]              Wrong timestamp format: $tstamp - cannot convert timestamp\n";
	}
	return "NULL";
    }
    
    return $sTime;

} # timestamp2String() ...


# ---------------------------------------------------------------------------
#
# extract the sinle log entries from a result row (the current $row):
# argument: $transLog
# returns: hash of log entries (key=value)
#
# ---------------------------------------------------------------------------
sub getTransLogEntries {

    my %logEntries;

    my @logTokens = split(/,/, $_[0]);
    my $token;
    foreach $token (@logTokens) {

	#print "[updateDB]              extracted from log: <$token>; ";

	my @entryParts = split(/=/, $token);

	#print "split into ".(scalar @entryParts)." parts.\n";
 
	if ($entryParts[0] ne "") {

	    # we found something meaningful:
	    my $entryName = $entryParts[0];
	    my $entryVal;

	    # check wether we splitted to many '=' (might be contained in the
	    # entry's value as well (such as cert subjects ...)
	    my $numEP = scalar(@entryParts);

	    if ($numEP == 2) {
		# exactly what we expect
		$entryVal = $entryParts[1];
	    }
	    elsif ($numEP > 2) {
		my $maxEPindex = $numEP - 1;
		# value was splitted because it contained '=', rebuild it:
		$entryVal = $entryParts[1]; #first part
		my $valPart;
		foreach $valPart (@entryParts[2 .. $maxEPindex]) {
		    $entryVal = $entryVal."=".$valPart; # additional parts
		}
	    }
	    else {
		# only one part, interpret it as a boolean value
		$entryVal = "";
	    }

	    # add the entry to the hash:
	    $logEntries{$entryName} = $entryVal;
	}

    }

    # transaction log already printed above, this was for testing if it is
    # parsed correctly ...
    
    if ($verboseOutput) {
	my $logKey;
	foreach $logKey (keys(%logEntries)) {
	    print "[updateDB]              from log: $logKey=<$logEntries{$logKey}>\n";
	}
    }

    return %logEntries;

} # getTransLogEntries() ...



# ---------------------------------------------------------------------------
#
# verify if a specific table exists
# arguments: $dbHandle, $table_name
# returns: 1 if exists, 0 if not or if an error occured.
#
# ---------------------------------------------------------------------------
sub existsTable() {

    my $dbh = $_[0];
    my $table = $_[1];

    my $dbQuery = "DESCRIBE ".$table;

    my $sth;

    eval { $sth = $dbh->prepare($dbQuery) };
    if ($@) {
	return 0; # error
    }

    eval { $sth->execute() };
    if ($@) {
	return 0; # error
    }

    return 1;  # exists

} # getColumnType() ...




# ---------------------------------------------------------------------------
#
# get a specific column type from the table description
# arguments: $dbHandle, $table_name, $field_name
# returns: $column_type, otherwise $QUERY_ERROR or $QUERY_NOT_FOUND
#
# ---------------------------------------------------------------------------
sub getColumnType() {

    my $dbh = $_[0];
    my $table = $_[1];
    my $field = $_[2];

    my $colType = $QUERY_NOT_FOUND; # default

    my $dbQuery = "DESCRIBE ".$table;

    if ($verboseOutput) {
	print "[updateDB] Getting table description: Query = $dbQuery\n";
    }

    my $sth;

    eval { $sth = $dbh->prepare($dbQuery) };
    if ($@) {
	print STDERR "[updateDB] Query preparation failed! Error: $@\n";
	return $QUERY_ERROR; # error
    }

    eval { $sth->execute() };
    if ($@) {
	print STDERR "[updateDB] Query execution failed! Error: $@\n";
	return $QUERY_ERROR; # error
    }


    while ($row = $sth->fetchrow_hashref()) {

	if (exists($row->{'Field'}) && exists($row->{'Type'})
	    && $row->{'Field'} eq $field) {

	    # found the right clolumn
	    $colType = $row->{'Type'};
	}

    } # while ($row ...

    return $colType;  # $QUERY_NOT_FOUND if not found

} # getColumnType() ...



# ---------------------------------------------------------------------------
#
# drop the primary key of a specific table
# arguments: $dbHandle, $table_name
# returns: $QUERY_ERROR, $QUERY_OK
#
# ---------------------------------------------------------------------------
sub dropPrimaryKey() {

    my $dbh = $_[0];
    my $table = $_[1];

    my $dbQuery = "ALTER TABLE ".$table." DROP PRIMARY KEY";

    print "[updateDB] Dropping primary key of table '$table'.\n";

    my $sth;

    eval { $sth = $dbh->prepare($dbQuery) };
    if ($@) {
	print STDERR "[updateDB] Query preparation failed! Error: $@\n";
	return $QUERY_ERROR; # error
    }

    eval { $sth->execute() };
    if ($@) {
	print STDERR "[updateDB] Query execution failed! Error: $@\n";
	return $QUERY_ERROR; # error
    }

    return $QUERY_OK; 
} # dropPrimaryKey() ...





# ---------------------------------------------------------------------------
#
# adding the primary key of a specific table
# arguments: $dbHandle, $table_name, $primaryKey
# returns: $QUERY_ERROR, $QUERY_OK
#
# ---------------------------------------------------------------------------
sub addPrimaryKey() {

    my $dbh = $_[0];
    my $table = $_[1];
    my $primaryKey = $_[2];

    my $dbQuery = "ALTER TABLE ".$table." ADD PRIMARY KEY (".$primaryKey.")";

    print "[updateDB] Adding primary key to table '$table': $primaryKey\n";

    my $sth;

    eval { $sth = $dbh->prepare($dbQuery) };
    if ($@) {
	print STDERR "[updateDB] Query preparation failed! Error: $@\n";
	return $QUERY_ERROR; # error
    }

    eval { $sth->execute() };
    if ($@) {
	print STDERR "[updateDB] Query execution failed! Error: $@\n";
	return $QUERY_ERROR; # error
    }

    return $QUERY_OK; 
} # addPrimaryKey() ...



# ---------------------------------------------------------------------------
#
# create a new table
# arguments: $dbHandle, $table_name, $table_specification
# returns: $QUERY_ERROR, $QUERY_OK
#
# ---------------------------------------------------------------------------
sub createTable() {

    my $dbh = $_[0];
    my $table = $_[1];
    my $tableSpec = $_[2];

    my $dbQuery = "CREATE TABLE ".$table." ".$tableSpec;

    print "[updateDB] Creating table '$table': Query = $dbQuery\n";

    my $sth;

    eval { $sth = $dbh->prepare($dbQuery) };
    if ($@) {
	print STDERR "[updateDB] Query preparation failed! Error: $@\n";
	return $QUERY_ERROR; # error
    }

    eval { $sth->execute() };
    if ($@) {
	print STDERR "[updateDB] Query execution failed! Error: $@\n";
	return $QUERY_ERROR; # error
    }

    return $QUERY_OK; 
} # createTable() ...



# ---------------------------------------------------------------------------
#
# creating new transInInfo/transOutInfo entries for already existing
# transactions. IMPORTANT: do this only if transInLog/transOutLog have
# been newly created and did not exist before!!!
# argument: $hlrDBhandle, $trans_type ("In"/"Out")
# returns: $QUERY_ERROR, $QUERY_OK or $QUERY_NOT_FOUND
#
# ---------------------------------------------------------------------------
sub updateHlr_createTransInfoEntries() {

    my $dbh = $_[0];
    my $transType = $_[1];

    if ($transType ne "In" && $transType ne "Out") {
	print STDERR "[updateDB] Cannot create trans(In/Out)Info entries: Unknown transaction\n";
	print STDERR "[updateDB] type '$transType'. Has to be 'In' or 'Out'.\n";
	return $QUERY_ERROR;
    }

    # source tables:
    my $transTable = "trans_".lc($transType);
    my $transLogTable = "trans".$transType."Log";
    # destination table:
    my $transInfoTable = "trans".$transType."Info";

    # +------------------+--------------+------+-----+---------+-------+
    # | Field            | Type         | Null | Key | Default | Extra |
    # +------------------+--------------+------+-----+---------+-------+
    # | dgJobId          | varchar(160) |      | PRI |         |       |
    # | tType            | varchar(160) | YES  |     | NULL    |       |
    # | userFqan         | varchar(160) | YES  |     | NULL    |       |
    # | userVO           | varchar(80)  | YES  |     | NULL    |       |
    # | jobStart         | int(11)      | YES  |     | NULL    |       |
    # | jobURCreation    | varchar(32)  | YES  |     | NULL    |       |
    # | resourceTimeZone | varchar(8)   | YES  |     | NULL    |       |
    # +------------------+--------------+------+-----+---------+-------+

    # get the single transactions to insert into $transInfoTable
    my $dbQuery = "SELECT "
	."$transTable.dg_jobid AS jobID, "
	."$transLogTable.log AS transLog, "
	."$transTable.tr_stamp AS transTimestamp "
	."FROM $transTable, $transLogTable "
	."WHERE $transTable.dg_jobid = $transLogTable.dgJobId";

    print "[updateDB] *** Creating '$transInfoTable' entries for existing transactions.\n";
    if ($verboseOutput) {
	print "[updateDB] Getting transactions from '$transTable' and '$transLogTable':\n";
	print "[updateDB] Query = $dbQuery\n";
    }

    my $sth;
    eval { $sth = $dbh->prepare($dbQuery) };
    if ($@) {
	print STDERR "[updateDB] Query preparation failed! Error: $@\n";
	return $QUERY_ERROR; # error
    }

    eval { $sth->execute() };
    if ($@) {
	print STDERR "[updateDB] Query execution failed! Error: $@\n";
	return $QUERY_ERROR; # error
    }

    while ($row = $sth->fetchrow_hashref()) {

	if ($verboseOutput) {
	    print "[updateDB] transaction: $transTable.dg_jobid=$row->{'jobID'}\n";
	    print "[updateDB]              $transLogTable.log=$row->{'transLog'}\n";
	    print "[updateDB]              $transTable.tr_stamp=$row->{'transTimestamp'}\n";
	}

	my %logEntries = &getTransLogEntries($row->{'transLog'});

	# contructing trans Info entry:

	my $dgJobId = $row->{'jobID'};

	my $tType = lc($transType);  # "in" or "out"

	my $userFqan = "";
	if (exists($logEntries{"userFqan"})) {
	    $userFqan = $logEntries{"userFqan"};
	}

	my $userVO = "";
	if ($userFqan =~ /^\/(\w*)\//) { # take from FQAN, more accurate
	    $userVO = $1;
	}
	elsif (exists($logEntries{"userVo"})) { # try to find elsewhere in log
	    $userVO = $logEntries{"userVo"};
	}

	my $jobStart = 0;
	if (exists($logEntries{"start"}) && ($logEntries{"start"} ne "")
	    && ($logEntries{"start"} =~ /^\d*$/)) {
	    # exists and is an integer:
	    $jobStart = $logEntries{"start"};
	}
	elsif (exists($logEntries{"CE_PRICE_TIME"}) 
	       && ($logEntries{"CE_PRICE_TIME"} ne "")
	       && ($logEntries{"CE_PRICE_TIME"} =~ /^\d*$/)) {
	    # take CE_PRICE_TIME instead:
	    $jobStart = $logEntries{"CE_PRICE_TIME"};
	}

	my $resourceTimeZone = "+0000"; # UTC
	if (exists($logEntries{"timeZone"})
	    && ($logEntries{"timeZone"} =~ /^(\+|-)\d{4}/)) {
	    # time zone in log:
	    $resourceTimeZone = $logEntries{"timeZone"};
	}

	my $jobURCreation = "";
	if (exists($logEntries{"urCreation"})
	    && ($logEntries{"urCreation"} ne "")) {
	    # ur creation exists in log:
	    $jobURCreation = $logEntries{"urCreation"};
	}
	else {
	    # doesn't exist, take transaction time stamp and transform to
	    # string:
	    $jobURCreation = &timestamp2String($row->{'transTimestamp'},
					       $resourceTimeZone);
	}

	#print "GOT: dgJobId = $dgJobId\n";
	#print "GOT: tType = $tType\n";
	#print "GOT: userFqan = $userFqan\n";
	#print "GOT: userVO = $userVO\n";
	#print "GOT: jobStart = $jobStart\n";
	#print "GOT: jobURCreation = $jobURCreation\n";
	#print "GOT: resourceTimeZone = $resourceTimeZone\n";

	my $insertDbQuery = "INSERT INTO $transInfoTable VALUES ("
	    ."\"$dgJobId\", \"$tType\", \"$userFqan\", \"$userVO\", "
	    ."$jobStart, \"$jobURCreation\", \"$resourceTimeZone\")";

	if ($verboseOutput) {
	    print "[updateDB] Inserting $transInfoTable entry: Query = $insertDbQuery\n";
	}

	my $insertSth;
	eval { $insertSth = $dbh->prepare($insertDbQuery) };
	if ($@) {
	    print STDERR "[updateDB] Query preparation failed! Error: $@\n";
	    return $QUERY_ERROR; # error
	}

	eval { $insertSth->execute() };
	if ($@) {
	    print STDERR "[updateDB] Query execution failed! Error: $@\n";
	    return $QUERY_ERROR; # error
	}

    } # while ($row ...

    return $QUERY_OK; # ok

} # updateHlr_createTransInfoEntries() ...




# ---------------------------------------------------------------------------
#
# creating new transInLog/transOutLog entries for already existing
# transactions. IMPORTANT: do this only if the version is not yet present.
# argument: $hlrDBhandle, $trans_type ("In"/"Out")
# returns: $QUERY_ERROR, $QUERY_OK or $QUERY_NOT_FOUND
#
# ---------------------------------------------------------------------------
sub updateHlr_addVersionToTransLogEntries() {

    my $dbh = $_[0];
    my $transType = $_[1];

    if ($transType ne "In" && $transType ne "Out") {
	print STDERR "[updateDB] Cannot create trans(In/Out)Info entries: Unknown transaction\n";
	print STDERR "[updateDB] type '$transType'. Has to be 'In' or 'Out'.\n";
	return $QUERY_ERROR;
    }

    my $engineVersionParam = "atmEngineVersion";
    my $engineVersionValue = "edg";
    my $clientVersionParam = "atmClientVersion";
    my $clientVersionValue = "edg";

    # source tables:
    my $transLogTable = "trans".$transType."Log";


    # get the single transactions from $transLogTable
    my $dbQuery = "SELECT "
	."dgJobId AS jobID, "
	."log AS transLog "
	."FROM $transLogTable";

    print "[updateDB] *** Checking '$transLogTable' entries for 'atmEngine/ClientVersion'.\n";
    if ($verboseOutput) {
	print "[updateDB] Getting transactions from '$transLogTable':\n";
	print "[updateDB] Query = $dbQuery\n";
    }

    my $sth;
    eval { $sth = $dbh->prepare($dbQuery) };
    if ($@) {
	print STDERR "[updateDB] Query preparation failed! Error: $@\n";
	return $QUERY_ERROR; # error
    }

    eval { $sth->execute() };
    if ($@) {
	print STDERR "[updateDB] Query execution failed! Error: $@\n";
	return $QUERY_ERROR; # error
    }

    while ($row = $sth->fetchrow_hashref()) {

	my $logUpdate = "";

	if ($verboseOutput) {
	    print "[updateDB] transaction: dgJobId=$row->{'jobID'}\n";
	    print "[updateDB]              log=$row->{'transLog'}\n";
	}

	my %logEntries = &getTransLogEntries($row->{'transLog'});

	if (exists($logEntries{$engineVersionParam})) {
	    # exists, don't add
	    if ($verboseOutput) {
		print "[updateDB] found in log: $engineVersionParam="
		    .$logEntries{$engineVersionParam}."\n";
	    }
	}
	else {
	    # doesn't exist, add the default defined above:
	    if ($logUpdate ne "") {
		$logUpdate = $logUpdate.",";
	    }
	    $logUpdate = $logUpdate.$engineVersionParam."="
		.$engineVersionValue;
	    if ($verboseOutput) {
		print "[updateDB] adding to log: $engineVersionParam="
		    .$engineVersionValue."\n";
	    }
	}

	if (exists($logEntries{$clientVersionParam})) {
	    # exists, don't add
	    if ($verboseOutput) {
		print "[updateDB] found in log: $clientVersionParam="
		    .$logEntries{$clientVersionParam}."\n";
	    }
	}
	else {
	    # doesn't exist, add the default defined above:
	    if ($logUpdate ne "") {
		$logUpdate = $logUpdate.",";
	    }
	    $logUpdate = $logUpdate.$clientVersionParam."="
		.$clientVersionValue;
	    if ($verboseOutput) {
		print "[updateDB] adding to log: $clientVersionParam="
		    .$clientVersionValue."\n";
	    }
	}


	if ($logUpdate ne "") {

	    # something to add:

	    my $updateDbQuery = "UPDATE $transLogTable SET log=\""
		.$row->{'transLog'}.",".$logUpdate."\" WHERE dgJobId=\""
		.$row->{'jobID'}."\"";

	    if ($verboseOutput) {
		print "[updateDB] Updating $transLogTable entry: Query = $updateDbQuery\n";
	    }

	    my $updateSth;
	    eval { $updateSth = $dbh->prepare($updateDbQuery) };
	    if ($@) {
		print STDERR "[updateDB] Query preparation failed! Error: $@\n";
		return $QUERY_ERROR; # error
	    }

	    eval { $updateSth->execute() };
	    if ($@) {
		print STDERR "[updateDB] Query execution failed! Error: $@\n";
		return $QUERY_ERROR; # error
	    }
	} # if ($logUpdate ne "") ...

    } # while ($row ...

    return $QUERY_OK; # ok

} # updateHlr_addVersionToTransLogEntries() ...




# ---------------------------------------------------------------------------
#
# update (if necessary) a specific column type from the table.
# Note: if $newColumnType is an empty string, the column type will be modified
#       without further check if necessary!
# arguments: $dbHandle, $table_name, $column_name, $newColumnType,
#            $modifyString
# returns: $QUERY_ERROR, $QUERY_OK or $QUERY_NOT_FOUND
#
# ---------------------------------------------------------------------------
sub updateColumnType() {

    my $dbh = $_[0];
    my $table = $_[1];
    my $column = $_[2];
    my $newColType = $_[3];
    my $modifyString = $_[4];

    my $oldColType = &getColumnType($dbh, $table, $column);

    if ($oldColType eq $QUERY_ERROR) {
	return $QUERY_ERROR;  # error
    }
    elsif ($oldColType eq $QUERY_NOT_FOUND) {
	print STDERR "[updateDB] Warning, no column '$column' in table '$table'.\n";
	return $QUERY_NOT_FOUND;  # error
    }
    else {
	# found! check if it has to be updated ...
	if (lc($oldColType) eq lc($newColType)) {  # not case-sensitive

	    print "[updateDB] Column '$column' in table '$table' up-to-date.\n";
	}
	else {
	    # update required ...

	    my $dbQuery = "ALTER TABLE ".$table." MODIFY ".$column." "
		          .$modifyString;

	    if ($verboseOutput) {
		print "[updateDB] Column '$column' in table '$table' is of type: $oldColType\n";
		if ($newColType ne "") {
		    print "[updateDB] Should be: $newColType\n";
		}
	    }
	    print "[updateDB] Updating table '$table': Query = $dbQuery\n";

	    my $sth;

	    eval { $sth = $dbh->prepare($dbQuery) };
	    if ($@) {
		print STDERR "[updateDB] Query preparation failed! Error: $@\n";
		return $QUERY_ERROR; # error
	    }

	    eval { $sth->execute() };
	    if ($@) {
		print STDERR "[updateDB] Query execution failed! Error: $@\n";
		return $QUERY_ERROR; # error
	    }
	}
    }

    return $QUERY_OK; 
} # updateColumnType() ...




# ---------------------------------------------------------------------------
#
# update (if necessary) a specific column type from the table
# for a column that is used as primary key as well, this involes redefining
# the primary key and should be used with care!!!!
# This method has to be used in case the column that is to be altered is part
# of the primary key and the current definition of the key doesn't allow the
# change. Example:
# current: acl VARCHAR(160),
#          PRIMARY KEY (acl)
# shall become: acl BLOB,
#               PRIMARY KEY (acl(255))
# (blobs cannot be primary keys if not as an index of fixed length)
#
# arguments: $dbHandle, $table_name, $column_name, $newColumnType,
#            $modifyString, $newPrimaryKey
# returns: $QUERY_ERROR, $QUERY_OK or $QUERY_NOT_FOUND
#
# ---------------------------------------------------------------------------
sub updateColumnTypeAndPrimaryKey() {

    my $dbh = $_[0];
    my $table = $_[1];
    my $column = $_[2];
    my $newColType = $_[3];
    my $modifyString = $_[4];
    my $newPrimaryKey = $_[5];

    # first check if we really have to do something:
    my $oldColType = &getColumnType($dbh, $table, $column);

    if ($oldColType eq $QUERY_ERROR) {
	return $QUERY_ERROR;  # error
    }
    elsif ($oldColType eq $QUERY_NOT_FOUND) {
	print STDERR "[updateDB] Warning, no column '$column' in table '$table'.\n";
	return $QUERY_NOT_FOUND;  # error
    }
    else {
	# found! check if it has to be updated ...
	if (lc($oldColType) ne lc($newColType)) { # not case-sensitive

	    # first drop the old key, otherwise the column cannot be updated
	    my $updateStatus = &dropPrimaryKey($dbh, $table);
	    if ($updateStatus ne $QUERY_OK) {
		return $updateStatus;
	    }

	    # update the column
	    $updateStatus = &updateColumnType($dbh, $table, $column,
					      $newColType, $modifyString);
	    if ($updateStatus ne $QUERY_OK) {
		# problem: already dropped the key, but cannot modify column!
		print STDERR "[updateDB] SEVERE ERROR:\n";
		print STDERR "[updateDB] Could not update the column '$column' in table '$table'!\n";
		print STDERR "[updateDB] Primary key has been dropped; fix by hand!!!\n";
		return $QUERY_ERROR;
	    }

	    # update primary key:
	    $updateStatus = &addPrimaryKey($dbh, $table, $newPrimaryKey);
	    if ($updateStatus ne $QUERY_OK) {
		# problem: old key dropped, new key cannot be added!
		print STDERR "[updateDB] SEVERE ERROR:\n";
		print STDERR "[updateDB] Could not add the primary key for table '$table': $newPrimaryKey!\n";
		print STDERR "[updateDB] Column has no primary key; fix by hand!!!\n";
		return $QUERY_ERROR;
	    }
	} # if (oldCol... ne newCol...
	else {
	    print "[updateDB] Column '$column' in table '$table' up-to-date.\n";
	}
    }

    return $QUERY_OK; 
} # updateColumnTypeAndPrimaryKey() ...




# ---------------------------------------------------------------------------
#
# updating table 'trans_queue' in database 'hlr_tmp'
# argument: none
#
# ---------------------------------------------------------------------------
sub updateHlrTmp_trans_queue() {

    my $retVal = 0; # ok

    my $updateStatus;

    print "[updateDB] *** Updating table 'trans_queue':\n";

    # from_cert_subject:
    $updateStatus = &updateColumnType($hlrTmpDbh, "trans_queue",
				      "from_cert_subject", "BLOB",
				      "BLOB NOT NULL default ''");
    if ($updateStatus ne $QUERY_OK) {
	$retVal = 1; # error
    }


    # to_cert_subject:
    $updateStatus = &updateColumnType($hlrTmpDbh, "trans_queue",
				      "to_cert_subject", "BLOB",
				      "BLOB NOT NULL default ''");
    if ($updateStatus ne $QUERY_OK) {
	$retVal = 1; # error
    }


    # type:
    $updateStatus = &updateColumnType($hlrTmpDbh, "trans_queue",
				      "type", "VARCHAR(32)",
				      "VARCHAR(32) NOT NULL default ''");
    if ($updateStatus ne $QUERY_OK) {
	$retVal = 1; # error
    }

    return $retVal;  # ok?

} # updateHlrTmp_trans_queue() ...




# ---------------------------------------------------------------------------
#
# updating table 'hlrAdmin' in database 'hlr'
# argument: none
#
# ---------------------------------------------------------------------------
sub updateHlr_hlrAdmin() {

    my $retVal = 0; # ok

    my $updateStatus;

    print "[updateDB] *** Updating table 'hlrAdmin':\n";


    # acl:
    $updateStatus = &updateColumnTypeAndPrimaryKey($hlrDbh, "hlrAdmin",
						   "acl", "BLOB",
						   "BLOB NOT NULL",
						   "acl(255)");
    if ($updateStatus ne $QUERY_OK) {
	$retVal = 1; # error
    }

    return $retVal;  # ok?

} # updateHlr_hlrAdmin() ...




# ---------------------------------------------------------------------------
#
# updating table 'voAdmin' in database 'hlr'
# argument: none
#
# ---------------------------------------------------------------------------
sub updateHlr_voAdmin() {

    my $retVal = 0; # ok

    my $updateStatus;

    print "[updateDB] *** Updating table 'voAdmin':\n";


    # acl:
    $updateStatus = &updateColumnTypeAndPrimaryKey($hlrDbh, "voAdmin",
						   "acl", "BLOB",
						   "BLOB NOT NULL",
						   "vo_id, acl(255)");
    if ($updateStatus ne $QUERY_OK) {
	$retVal = 1; # error
    }

    return $retVal;  # ok?

} # updateHlr_voAdmin() ...




# ---------------------------------------------------------------------------
#
# updating table 'groupAdmin' in database 'hlr'
# argument: none
#
# ---------------------------------------------------------------------------
sub updateHlr_groupAdmin() {

    my $retVal = 0; # ok

    my $updateStatus;

    if (&existsTable($hlrDbh, "groupAdmin")) {

	# exists, update:
	print "[updateDB] *** Updating table 'groupAdmin':\n";


	# gid:
	$updateStatus = &updateColumnType($hlrDbh, "groupAdmin",
					  "gid",
					  "",  # don't check current type (same
					       # without NOT NULL)
					  "VARCHAR(128) NOT NULL");
	if ($updateStatus ne $QUERY_OK) {
	    $retVal = 1; # error
	}

	# acl:
	$updateStatus = &updateColumnTypeAndPrimaryKey($hlrDbh, "groupAdmin",
						       "acl", "BLOB",
						       "BLOB NOT NULL",
						       "gid, acl(255)");
	if ($updateStatus ne $QUERY_OK) {
	    $retVal = 1; # error
	}
    }
    else {
	# doesn't exist, create it!
	print "[updateDB] *** Creating table 'groupAdmin':\n";

	my $tableSpec = "( "
	    ."gid varchar(128) NOT NULL, "
	    ."acl blob NOT NULL, "
	    ."PRIMARY KEY (gid,acl(255)) "
	    .") TYPE=MyISAM";

	$updateStatus = &createTable($hlrDbh, "groupAdmin", $tableSpec);
	if ($updateStatus ne $QUERY_OK) {
	    $retVal = 1; # error
	}

    }

    return $retVal;  # ok?

} # updateHlr_groupAdmin() ...





# ---------------------------------------------------------------------------
#
# updating table 'jobAuth' in database 'hlr'
# argument: none
#
# ---------------------------------------------------------------------------
sub updateHlr_jobAuth() {

    my $retVal = 0; # ok

    my $updateStatus;

    print "[updateDB] *** Updating table 'jobAuth':\n";

    # usrCert:
    $updateStatus = &updateColumnType($hlrDbh, "jobAuth",
				      "usrCert", "BLOB",
				      "BLOB NOT NULL");
    if ($updateStatus ne $QUERY_OK) {
	$retVal = 1; # error
    }

    return $retVal;  # ok?

} # updateHlr_jobAuth() ...





# ---------------------------------------------------------------------------
#
# updating table 'acctdesc' in database 'hlr'
# argument: none
#
# ---------------------------------------------------------------------------
sub updateHlr_acctdesc() {

    my $retVal = 0; # ok

    my $updateStatus;

    print "[updateDB] *** Updating table 'acctdesc':\n";

    # usrCert:
    $updateStatus = &updateColumnType($hlrDbh, "acctdesc",
				      "cert_subject", "BLOB",
				      "BLOB DEFAULT NULL");
    if ($updateStatus ne $QUERY_OK) {
	$retVal = 1; # error
    }

    return $retVal;  # ok?

} # updateHlr_acctdesc() ...






# ---------------------------------------------------------------------------
#
# updating table 'transInInfo' in database 'hlr'
# argument: none
#
# ---------------------------------------------------------------------------
sub updateHlr_transInInfo() {

    my $retVal = 0; # ok

    my $updateStatus;

    if (&existsTable($hlrDbh, "transInInfo")) {
	# exists:
	print "[updateDB] *** Table 'transInInfo' exists; nothing to do.\n";
    }
    else {
	# doesn't exist, create it!
	print "[updateDB] *** Creating table 'transInInfo':\n";

	my $tableSpec = "( "
	    ."dgJobId varchar(160) NOT NULL, "
	    ."tType varchar(160), "
	    ."userFqan varchar(160), "
	    ."userVO varchar(80), "
	    ."jobStart int(11), "
	    ."jobURCreation varchar(32), "
	    ."resourceTimeZone varchar(8), "
	    ."PRIMARY KEY (dgJobId) "
	    .") TYPE=MyISAM";

	$updateStatus = &createTable($hlrDbh, "transInInfo", $tableSpec);
	if ($updateStatus ne $QUERY_OK) {
	    $retVal = 1; # error
	}

	# now create entries for existing transactions:
	$updateStatus = &updateHlr_createTransInfoEntries($hlrDbh, "In");
	if ($updateStatus ne $QUERY_OK) {
	    $retVal = 2; # error
	}

    }

    return $retVal;  # ok?

} # updateHlr_transInInfo() ...




# ---------------------------------------------------------------------------
#
# updating table 'transOutInfo' in database 'hlr'
# argument: none
#
# ---------------------------------------------------------------------------
sub updateHlr_transOutInfo() {

    my $retVal = 0; # ok

    my $updateStatus;

    if (&existsTable($hlrDbh, "transOutInfo")) {
	# exists:
	print "[updateDB] *** Table 'transOutInfo' exists; nothing to do.\n";
    }
    else {
	# doesn't exist, create it!
	print "[updateDB] *** Creating table 'transOutInfo':\n";

	my $tableSpec = "( "
	    ."dgJobId varchar(160) NOT NULL, "
	    ."tType varchar(160), "
	    ."userFqan varchar(160), "
	    ."userVO varchar(80), "
	    ."jobStart int(11), "
	    ."jobURCreation varchar(32), "
	    ."resourceTimeZone varchar(8), "
	    ."PRIMARY KEY (dgJobId) "
	    .") TYPE=MyISAM";

	$updateStatus = &createTable($hlrDbh, "transOutInfo", $tableSpec);
	if ($updateStatus ne $QUERY_OK) {
	    $retVal = 1; # error
	}

	# now create entries for existing transactions:
	$updateStatus = &updateHlr_createTransInfoEntries($hlrDbh, "Out");
	if ($updateStatus ne $QUERY_OK) {
	    $retVal = 2; # error
	}

    }

    return $retVal;  # ok?

} # updateHlr_transOutInfo() ...


# ---------------------------------------------------------------------------
#
# updating table 'transInLog' in database 'hlr'
# argument: none
#
# ---------------------------------------------------------------------------
sub updateHlr_transInLog() {

    my $retVal = 0; # ok

    my $updateStatus;

    # now create entries for existing transactions:
    $updateStatus = &updateHlr_addVersionToTransLogEntries($hlrDbh, "In");
    if ($updateStatus ne $QUERY_OK) {
	$retVal = 2; # error
    }

    return $retVal;  # ok?

} # updateHlr_transInLog() ...


# ---------------------------------------------------------------------------
#
# updating table 'transOutLog' in database 'hlr'
# argument: none
#
# ---------------------------------------------------------------------------
sub updateHlr_transOutLog() {

    my $retVal = 0; # ok

    my $updateStatus;

    # now create entries for existing transactions:
    $updateStatus = &updateHlr_addVersionToTransLogEntries($hlrDbh, "Out");
    if ($updateStatus ne $QUERY_OK) {
	$retVal = 2; # error
    }

    return $retVal;  # ok?

} # updateHlr_transOutLog() ...
