#!/usr/bin/perl
# This script reads an edg-mkgridmap-like configuration file
# and uses edg-mkgridmap for every "group" entry in the configuration file.
# (the queried servers can use different protocols, such as HTTPS, LDAP or
#  VOMS to export the information on members)
# The created gridmap-file is parsed to retrieve the user information and
# create the respective HLR accounts.
# Authors: Rosario M. Piro and Andrea Guarise; {piro,guarise}@to.infn.it

use strict;

use Getopt::Std;

my $prgname = "glite-dgas-hlr_userAccountImport.pl";
                                                                                
my %opts;
my $forceFlag = "";
                                                                                
if ( $ENV{GLITE_LOCATION} eq "" ) {
        $ENV{GLITE_LOCATION}="/opt/glite/";
}

if ( $ENV{EDG_LOCATION} eq "" ) {
        $ENV{EDG_LOCATION}="/opt/edg/";
}



my $conffile = "$ENV{GLITE_LOCATION}/etc/glite-dgas-hlr_mkgridmap_userAccountImport.conf";
my $tmpconffile = "/tmp/tmpmkgridmap_conffile";
my $multiVOuserLogfile = "";



# -------------------
# getting options ...
# -------------------
                                                                                
getopts('fhc:g:l:', \%opts);
                                                                                
if ( $opts{h} eq "1" ) {
	print "$prgname imports HLR user accounts from multiple VOMS or LDAP servers.\n";
	print "Authors: A. Guarise, R. Piro\n\n";
        print "Usage:\n";
        print "$prgname [OPTIONS]\n";
        print "\nWhere OPTIONS are:\n\n";
        print "-f         Force flag, if specified existing accounts are overwritten.\n";
        print "-c <file>  specifies the mkgridmap configuration file.\n";
        print "           Default: $conffile\n";
	print "-l <file>  specifies the log file that will contain (only!) information on\n";
        print "           multi VO users. Normal log information is always written to stdout!\n";
        print "           Default: stdout\n";
        exit(1);
}
                                                                                
if ( $opts{f} eq "1" ) {
        $forceFlag = "-F";
};
                                                                                
if ( $opts{c} ne "" ) {
        $conffile = $opts{c};
}
                                                                                
if ( $opts{l} ne "" ) {
        $multiVOuserLogfile = $opts{l};
}


#
# initializing logfiles, if necessary:
# 
                                                                               
if ( $multiVOuserLogfile ne "" ) {
	unless ( open (MVOLOG, ">$multiVOuserLogfile") ) {
                print STDERR "Error: cannot create log file for multi VO users, using stdout instead!\n";
		*MVOLOG = *STDOUT;
        }
	else {
		print MVOLOG "# multi VO user logfile created by $prgname\n# NOTE: users may appear more then once in this list!\n# Format: |<user name>|<certificate subject>|VO(account):<vo name>|VO(other):<vo name>|\n";
	}
}
else {
	# not specified, use stdout!
	*MVOLOG = *STDOUT;
}

# will contain the VO name of a user account on the HLR (to determine whether
# the user we are trying to import is multi VO or not):
my $existingUserVO = "";


# ------------------------------
# parsing of configuration file:
# ------------------------------

my $line = "";

unless ( -e $conffile ) {
	print STDERR "Error: mkgridmap configuration file not found!\n";
	exit(1);
}

unless ( -r $conffile ) {
	print STDERR "Error: mkgridmap configuration file not readable!\n";
	exit(1);
}
         
unless ( open(IN, "< $conffile") ) {
	print STDERR "Error: cannot open mkgridmap configuration file!\n";
        exit(1);
}




# initializing accountInfo structure:
my %accountInfo = (
                dn => "",
                mail => "",
                description => ""
		);
my $dn;
my $cn;
my $group;
my $mail;
my $uid;
my $vo;


# counters for added HLR accounts, multi VO users and single VO users:
my $addedUsersCounter = 0;
my $multiVOuserCounter = 0;
my $singleVOuserCounter = 0;
my $thisVO_addedUsersCounter = 0;
my $thisVO_multiVOuserCounter = 0;
my $thisVO_singleVOuserCounter = 0;

# reading the single lines
while ($line = <IN>) {
	chomp($line);

	$line || next;

	next if ($line =~ /^\s*\#+/);

	# the following is done only for group entries:
	if ($line =~ /^\s*group(.*)/) {

		print "# processing line: $line\n";

		# writing the line as a temporary configuration file
		# for edg-mkgridmap ...
		my $cmd = "echo '# written by $prgname' > $tmpconffile ; echo $line >> $tmpconffile";
		my $status = system($cmd);

		if ( $status ne 0 ) {
			print STDERR "Error: cannot write temporary conf. file $tmpconffile\n";
			exit(1);
		}

		# running edg-mkgridmap with the temporary configuration file:
		if (&runMkgridmap != 0)
		{
			print STDERR "Error: error while running edg-mkgridmap to process the line '$line' of the configuration file!\n";
			# exit(1);
			next; # don't stop, try next VO instead!
		}

		# parsing temporary gridmap-file and creating user accounts:
		parseTmpGridMapFile();

		# statistics for this VO:
		print "VO '$vo': total users: ".($thisVO_singleVOuserCounter+$thisVO_multiVOuserCounter).", multi VO users: $thisVO_multiVOuserCounter, single VO users: $thisVO_singleVOuserCounter, new accounts: $thisVO_addedUsersCounter.\n";

	} # if line begins with group

} # while ...


# final statistics for all VOs:
print "ALL VOs: total users: ".($singleVOuserCounter+$multiVOuserCounter).", multi VO users: $multiVOuserCounter, single VO users: $singleVOuserCounter, new accounts: $addedUsersCounter.\n";

# ------------
# subroutines:
# ------------

sub runMkgridmap() {

        # first resetting temporary grid-mapfile. If the current
        # VO doesn't have valid users the file will not be overwritten
        # by mkgridmap and the previous VO would be processed a second
        # time by this script (if the temporary grid-mapfile
        # does already exist from a previous run).
        my $cmd = "rm -f /tmp/tmpgridmap; touch /tmp/tmpgridmap";
        my $status = system($cmd);
        if ( $status ne 0 ) {
		print STDERR "Error: cannot write temporary grid-mapfile /tmp/tmpgridmap\n";
                exit(1);
        }
                                                                                
        $cmd = "$ENV{EDG_LOCATION}/sbin/edg-mkgridmap --conf=$tmpconffile --output=/tmp/tmpgridmap";

	print "$cmd\n";

        my $ret = system($cmd);
        return $ret;
}



sub parseTmpGridMapFile() {

	# This subroutine parses the temporary gridmap-file and creates
	# the necessary user accounts:

	$thisVO_addedUsersCounter = 0;
	$thisVO_multiVOuserCounter = 0;
	$thisVO_singleVOuserCounter = 0;

	unless ( open(FILE, "/tmp/tmpgridmap") ) {
		print STDERR "Error: cannot open temporary gridmap-file /tmp/tmpgridmap";
		exit(1);
	}

	while(<FILE>) {
		print "\nLINE:$_";

        	$dn = $_;
        	$dn =~ /"(.*)" \.(.*)/;
        	$dn = $1;
        	$vo = $2;
		print "DN:$dn\n";
        	print "VO:$vo\n";

        	$mail =  "";
        	$mail = $dn;
        	$mail =~ /\/Email=(.*)\//;
        	$mail =$1;
        	if ($mail eq $dn)
        	{
                	$mail =~ /\/Email=(.*)/;
                	$mail = $1;
                	if ($mail eq $dn)
                	{
                        	$mail = "Unknown";
                	}
        	}
        	print "MAIL:$mail\n";

        	$cn = "";
        	$cn = $dn;
        	$cn =~ /\/CN=(.*)\//;
        	$cn =$1;
        	if ($cn eq $dn)
        	{
                	$cn =~ /\/CN=(.*)/;
                	$cn = $1;
        	}
        	print "CN:$cn\n";
        	#$group =$opts{G};
                                                                                
        	$accountInfo{dn}=$dn;
        	# the cert subject might contain a ' that is not recognized by MySQL if
        	# not preceeded by an escape character!
        	$accountInfo{dn}=~s/'/\\'/g;
        	$accountInfo{mail}=$mail;
        	$accountInfo{description}=$cn;
        	$accountInfo{description}=~s/'//g;
        	$accountInfo{vo_id}=$vo;
        	$group = "";
        	$group = $accountInfo{mail};
                                                                                
        	if ($group ne "Unknown")
        	{
                	$group =~ /(.*)@(.*)$/;
                	$group = $2;
        	}
                                                                                
        	# added in case the email is not specified
        	# since the addUser program requires a group name!
        	if ( $group eq "Unknown" )
        	{
                	$group = "group_not_specified";
        	}
                                                                                
        	# mail domains may be in different VOs, we therefore specify the
        	# VO in the group name, to distinguish them ...

        	$group = $vo."_".$group;
                                                                                
        	$accountInfo{gid}=$group;
        	$uid = $cn;
                                                                                
        	# if the uid is STILL empty (the cert subject eventually doesn't
        	# contain the field CN ...) we set a uid for an "not specified" user
        	if ( $uid eq "" )
        	{
                	$uid = $vo."_user_not_specified";
        	}
                                                                                
        	$uid =~ s/ //g;
        	$uid =~ s/'//g;
                                                                                
		# we do that later, when we are sure we add the user!
        	## make sure each userid exists only once:
        	#my $count = 2;
        	#my $newuid = $uid;
        	#while ( &userExists($newuid) == 0 )
        	#{
                #	$newuid = $uid . $count;
                #	$count++;
        	#}
        	#$uid = $newuid;
                #
        	#$accountInfo{uid}=$uid;
        	#my $aref = \%accountInfo;

        	if ( &certExists($accountInfo{dn}) == 0 ) {

                	#the dn  already exists!
			if ( $existingUserVO eq $vo ) {
				# account exists for same VO!
                		print "Account already exists for: '$accountInfo{description}' with subject '$accountInfo{dn}' and VO '$vo'\n";

				$thisVO_singleVOuserCounter++;
				$singleVOuserCounter++;
			}
			else {
				# multi VO user!!!
				print "Multi VO user; the user is already registered for the VO '$existingUserVO'!\n";
				print MVOLOG "|$accountInfo{description}|$accountInfo{dn}|VO(account):$existingUserVO|VO(other):$vo|\n";

				$thisVO_multiVOuserCounter++;
				$multiVOuserCounter++;
			}
        	}
        	else {

			# certificate not on HLR db: adding the user!

                        # first make sure the VO exists:
        	        if ( &voExists($accountInfo{vo_id}) != 0 ) {
                	        &createVO($accountInfo{vo_id});
        	        }

                        # now make sure the group exists:
                	if ( &groupExists($accountInfo{gid}) != 0 ) {
                        	&createGroup($accountInfo{vo_id},$accountInfo{gid});
                	}


                	# make sure each userid exists only once:
                	my $count = 2;
                	my $newuid = $uid;
                	while ( &userExists($newuid) == 0 )
                	{
                        	$newuid = $uid . $count;
                        	$count++;
                	}
                	$uid = $newuid;
			$accountInfo{uid}=$uid;

			my $aref = \%accountInfo;

                	if ( &createAccount($aref) == 0 ) {
				# new account!
                        	print "$addedUsersCounter\n";

				$thisVO_addedUsersCounter++;
                        	$addedUsersCounter++;
                        	$thisVO_singleVOuserCounter++;
                        	$singleVOuserCounter++;
                	}

        	} # else ... (user has to be added to DB).

	} # while FILE

} # sub parse...


sub userExists
{
        my $uid = $_[0];
        my $prefix = "$ENV{GLITE_LOCATION}/sbin/";
        my $cmd = "$prefix/glite_dgas_hlrQueryUser -u '$uid'";

	print "$cmd\n";

        my $ret = system($cmd);
        return $ret;
}


sub certExists
{
        my $cert = $_[0];
        my $prefix = "$ENV{GLITE_LOCATION}/sbin/";

        my $cmd = "$prefix/glite_dgas_hlrQueryUser -U -c \"$cert\"";

	print "$cmd\n";

        #my $ret = system($cmd);

	# we also get the output string, since we have to get the
	# VO with which a user is registered in order to determine
	# which users are multi VO users:

	my $retstring = `$cmd`; # stdout of $cmd goes to $retstring!

	my $ret = 0;
	if ( $retstring ne "" ) {
		$existingUserVO = "";
		chomp($retstring);
		print "$retstring\n";
		# we get the VO of the user to later compare it to the VO
		# for which we wanted to add the user:
		my @userInfo = split(/\|/, $retstring);
		$existingUserVO = @userInfo[6];
	}
	else {
		# no output -> cert doesn't exist!
		print "User account doesn't exist!\n";
		$ret = 1;
	} 

        return $ret;
}



sub createAccount()
{
        my %Info = %{$_[0]};
        my $prefix = "$ENV{GLITE_LOCATION}/sbin/";
        my $cmd = "$prefix/glite_dgas_hlrAddUser $forceFlag -u \"$Info{uid}\" --vo_id \"$vo\" -g '$group' -d \"$Info{description}\" -e '$Info{mail}' -c \"$Info{dn}\"";

	print "$cmd\n";

        my $ret = system($cmd);
        return $ret;
}



sub groupExists
{
        my $group = $_[0];
        my $prefix = "$ENV{GLITE_LOCATION}/sbin/";
        my $cmd = "$prefix/glite_dgas_hlrQueryGroup -g '$group'";

        print "$cmd\n";

        my $ret = system($cmd);
        return $ret;
}



sub createGroup()
{
        my $vo = $_[0];
        my $group = $_[1];
        my $prefix = "$ENV{GLITE_LOCATION}/sbin/";
        my $cmd ="$prefix/glite_dgas_hlrAddGroup $forceFlag --vo_id \"$vo\" -g '$group'";

        print "$cmd\n";

        return system($cmd);
}


sub voExists
{
        my $vo = $_[0];
        my $prefix = "$ENV{GLITE_LOCATION}/sbin/";
        my $cmd = "$prefix/glite_dgas_hlrQueryVO --vo_id \"$vo\"";

        print "$cmd\n";

        my $ret = system($cmd);
        return $ret;
}


sub createVO()
{
        my $vo = $_[0];
        my $prefix = "$ENV{GLITE_LOCATION}/sbin/";
        my $voDescription = "$line";
        my $cmd ="$prefix/glite_dgas_hlrAddVO $forceFlag --vo_id \"$vo\" -d '$voDescription' ";

        print "$cmd\n";

        return system($cmd);

}
